# Define some global variables needed for setup
# Derek Fujimoto
# Aug 2021

from os.path import join

# variables
__src__ = join('bfit', 'fitting')
__version__ = '4.22.2'

# keywords used to identify variables
KEYVARS = { 'B0'    : 'B0 Field (T)',
            'BIAS'  : 'Platform Bias (kV)',
            'CLFT'  : 'Cryo Lift Read (mm)',
            'CNT'   : 'Sample Counts',
            'DUR'   : 'Run Duration (s)',
            'ENRG'  : 'Impl. Energy (keV)',
            'LAS'   : 'Laser Power (V)',
            'NBMC'  : 'NBM Counts',
            'NBMR'  : 'NBM Rate (count/s)',
            'RATE'  : 'Sample Rate (count/s)',
            'RF'    : 'RF Level DAC',
            'RUN'   : 'Run Number',
            'TEMP'  : 'Temperature (K)',
            'TIME'  : 'Start Time',
            'YEAR'  : 'Year',
          }



# Model the fit results with a function
# Derek Fujimoto
# Nov 2019


from tkinter import *
from tkinter import ttk, messagebox
from functools import partial

import logging, re, os, warnings
import jax
import numpy as np
import jax.numpy as jnp
import pandas as pd
import bdata as bd

from bfit.global_variables import KEYVARS
from bfit.global_variables import KEYVARS
from bfit.gui.template_fit_popup import template_fit_popup
from bfit.gui.InputLine import InputLine

jax.config.update('jax_platform_name', 'cpu')
jax.config.update("jax_enable_x64", True)

# ========================================================================== #
class popup_fit_constraints(template_fit_popup):
    """
        Popup window for settings fit parameters according to a function
        
        bfit
        fittab
        logger
        
        constraints_are_set bool, if true, constraints should be in place
        defined             list of str, defined parameter names
        eqn                 list of str, equations for each defined parameter
        
        input_error         Bool, if true don't allow contrain to run
        
        label_new_var       Label, show new variables
        label_defined       Label, show which variables will be redefined
        
        new_par             list of list of str, new parameter found in each eqn
        new_par_unique      list of str, new parameters, sorted
        
        output_par_text     text, detected parameter names
        output_text         dict, keys: p0, blo, bhi, res, err, value: tkk.Text objects
       
        output_par_text_val string, contents of output_par_text
        output_text_val     dict of strings, contents of output_text
       
        parnames:           list, function inputs
        reserved_pars:      dict, define values in bdata that can be accessed
        win:                Toplevel
    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}
    window_title = 'Constrain parameters'
    
    # default values
    defaults = {'p0':       1,
                'blo':      -np.inf,
                'bhi':      np.inf,
                'res':      np.nan,
                'dres+':    np.nan,
                'dres-':    np.nan,
                'chi':      np.nan,
                'fixed':    False,
                'shared':   False,
                }
    
    # ====================================================================== #
    def __init__(self, bfit, fittab, input_fn_text=''):
        
        super().__init__(bfit, input_fn_text)
        
        # initialize
        self.defined = []
        self.eqn = []
        self.new_par = []
        self.new_par_unique = []
        self.fittab = fittab
        self.input_error = False
        self.constraints_are_set = False
        
    # ====================================================================== #
    def add_fn(self, data):
        """
            Make a lambda function to add to data
        """
        
        # trivial end
        if not self.constraints_are_set:
            data.constrained = {}
            return
        
        # get variables in decreasing order of length (no mistakes in replace)
        varlist = np.array(list(KEYVARS.keys()))
        varlist = varlist[np.argsort(list(map(len, varlist))[::-1])]
    
        # make functions
        fns = {}
        for defined, eqn, new_par in zip(self.defined, self.eqn, self.new_par):
            
            # find constant names in the string, replace with constant
            for var in varlist:
                if var in eqn:
                    value = data.get_values(KEYVARS[var])[0]
                    eqn = eqn.replace(var, str(value))
            
            new_par.sort()
            f = 'lambda {new_par} : {equation}'
            f = f.format(new_par=','.join(new_par), 
                         equation=eqn)
                                       
            # replace numpy functions with jax.numpy functions
            f = f.replace('np.', 'jnp.')
                                      
            # evaluate functions string to python handle
            fns[defined] = (eval(f), new_par)
            
        data.constrained = fns
        
    # ====================================================================== #
    def add_new_par(self, data):
        """
            Add constrained parameters to data.fitpar
        """
        
        # clean nan rows
        for p in self.new_par_unique:
            if p in data.fitpar.index and all(data.fitpar.loc[p, ['p0', 'blo', 'bhi']].isna()):
                data.fitpar.drop(index=p, inplace=True)
        
        # check if no changes to input list
        if not any((p not in data.fitpar.index for p in self.new_par_unique)):
            return 
            
        # check if constrained
        if not self.constraints_are_set:
            return
        
        # add new parameters, looking for present values
        cols = InputLine.columns
        new_fit_par = {}
        new_par_added = []
        
        # iterate parameter names
        for c in cols:
            
            # get values
            values = []
            for par in self.new_par_unique:
                if par in data.fitpar.index:                    
                    values.append(data.fitpar.loc[par, c])
                else:
                    if par not in new_par_added:
                        new_par_added.append(par)
                    values.append(self.defaults[c])
                    
            # set values
            new_fit_par[c] = values
            
        data.set_fitpar(pd.DataFrame(new_fit_par, 
                                     index=self.new_par_unique)
                        )

    # ====================================================================== #
    def do_after_parse(self, defined=None, eqn=None, new_par=None):
        """
            show inputs as readback and save
        """
        
        if defined and defined is not None:
        
            # check defined variables
            ncomp = self.bfit.fit_files.n_component.get()
            fn_name = self.bfit.fit_files.fit_function_title.get()
            par_names = self.bfit.fit_files.fitter.gen_param_names(fn_name, ncomp)
            s = [d if d else '<blank>' for d in sorted(defined)]
            s = [d if d in par_names else '%s [ERROR: Bad parameter]' % d for d in s]
            s = '\n'.join(s)
            
            # check for errors
            self.input_error = 'ERROR' in s
            
            # set label
            self.label_defined.config(text=s)
            
            # check new variables
            s = sorted(np.unique(np.concatenate(new_par)))
            s = [i for i in s if i and i not in self.parnames]
            
            # show input
            self.label_new_var.config(text='\n'.join(s))
        
        else:
            self.label_defined.config(text='')
            self.label_new_var.config(text='')
            defined = []
            eqn = []
            new_par = []
            self.input_error = False
        
        # save
        self.defined = defined
        self.eqn = eqn
        self.new_par = new_par
        
        try:
            self.new_par_unique = sorted(np.unique(np.concatenate(new_par)))
            self.new_par_unique = list(map(str, self.new_par_unique))
        except ValueError:
            self.new_par_unique = []
    
    # ====================================================================== #
    def do_return(self, *_):
        """
            Activated on press of return key
        """
        self.set_constraints()

    # ====================================================================== #
    def drop_unused_param(self, data):
        """
            Remove rows from fitdata.fitpar if not constrained new_par
        """
        # get function input names
        fit_files = self.bfit.fit_files
        ncomp = fit_files.n_component.get()
        fn_title = fit_files.fit_function_title.get()
        pnames = sorted(fit_files.fitter.gen_param_names(fn_title, ncomp))
        
        # check if constrained, get parnames
        if self.constraints_are_set:
            pnames.extend(self.new_par_unique)
        
        # drop unused
        data.drop_unused_param(pnames)
        
    # ====================================================================== #
    def set_constraints(self):
        """
            Set up constraining parameter functions
        """
        
        # check for input errors
        if self.input_error:
            msg = 'Input error'
            self.logger.exception(msg)
            messagebox.showerror('Error', msg)
            raise RuntimeError(msg)
        
        # no constraints: enable all lines
        if not self.defined:                
            self.constraints_are_set = False
            self.fittab.populate()
            self.cancel()
            return
            
        # check for missing equations
        if any([e=='' for e in self.eqn]):
            defi = self.defined[self.eqn.index('')]
            msg = 'Missing equation for {d}'.format(d=defi)
            self.logger.exception(msg)
            messagebox.showerror('Error', msg)
            raise RuntimeError(msg)
        
        # check for circular definitions
        for d, e in zip(self.defined, self.eqn):
            if d in e:
                msg = 'Circular parameter definitions not allowed:'+\
                      '\n{d} = f({d}) is not allowed'.format(d=d)
                messagebox.showerror('Error', msg)
                self.logger.exception(msg)
                raise RuntimeError(msg)
        
        # check for more circular definitions
        for d in self.parnames:
            for e in self.eqn:
                if d in e:
                    msg = 'Reserved parameter "{p}" cannot be used as an input'.format(p=d)
                    self.logger.exception(msg)
                    messagebox.showerror('Error', msg)
                    raise RuntimeError(msg)
        
        # set flag
        self.constraints_are_set = True
        
        # populate 
        self.fittab.populate()
        
        # close
        self.cancel()
    
    # ====================================================================== #
    def set_init_button_state(self, fline):
        """
            Set the state of the gui_param_buttons in fit_lines
        """
        if self.constraints_are_set:
            state = 'disabled'
        else:
            state = 'normal'
        
        fline.gui_param_button.config(state=state)
            
    # ====================================================================== #
    def show(self):
        """
            show window
        """
        
        # show base class
        if not hasattr(self, 'win') or not Toplevel.winfo_exists(self.win):
            super().show()
        
        # Keyword parameters
        key_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved variable names:\n\n'
        self.reserved_pars = KEYVARS
        
        keys = list(self.reserved_pars.keys())
        descr = [self.reserved_pars[k] for k in self.reserved_pars]
        maxk = max(list(map(len, keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        key_param_label = ttk.Label(key_param_frame, text=s, justify=LEFT)
        
        # fit parameter names 
        fit_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved function parameter names:\n\n'
        self.parnames = self.fittab.fitter.gen_param_names(
                                        self.fittab.fit_function_title.get(), 
                                        self.fittab.n_component.get())
        
        s += '\n'.join([k for k in sorted(self.parnames)]) 
        s += '\n'
        fit_param_label = ttk.Label(fit_param_frame, text=s, justify=LEFT)

        # module names 
        module_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved module names:\n\n'
        
        keys = list(self.modules.keys())
        descr = [self.modules[k] for k in self.modules]
        maxk = max(list(map(len, keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        modules_label = ttk.Label(module_frame, text=s, justify=LEFT)
        
        # Text entry
        self.entry_label['text'] = 'Enter one constraint equation per line.'+\
                                 '\nNon-reserved words are shared variables.'+\
                                 '\nEx: "1_T1 = a*np.exp(b*BIAS**0.5)+c"'                                 
                
        # detected new parameters
        frame_detected = ttk.Frame(self.right_frame)
        label_defined_title = ttk.Label(frame_detected, text='Constrained parameters')
        label_new_var_title = ttk.Label(frame_detected, text='New variables')
        
        self.label_new_var = ttk.Label(frame_detected, text='')
        self.label_defined = ttk.Label(frame_detected, text='')
                
        # add constrain button
        button_constrain = ttk.Button(self.right_frame, text='Constrain', 
                                      command=self.set_constraints)
                
        # gridding
        key_param_label.grid(column=0, row=0)
        fit_param_label.grid(column=0, row=0)
        modules_label.grid(column=0, row=0)
        
        key_param_frame.grid(column=0, row=0, rowspan=1, sticky='ew', padx=1, pady=1)
        module_frame.grid(column=0, row=1, sticky='ew', padx=1, pady=1, rowspan=2)
        fit_param_frame.grid(column=0, row=3, sticky='ewns', padx=1, pady=1)
        
        frame_detected.grid(column=0, row=4, sticky='ewn', padx=1, pady=1)
        frame_detected.grid_columnconfigure(0, weight=1)
        frame_detected.grid_columnconfigure(1, weight=1)
        label_defined_title.grid(column=0, row=0, sticky='n', padx=1, pady=1)
        label_new_var_title.grid(column=1, row=0, sticky='n', padx=1, pady=1)
        self.label_defined.grid( column=0, row=1, sticky='n', padx=1, pady=1)
        self.label_new_var.grid( column=1, row=1, sticky='n', padx=1, pady=1)
        
        
        self.right_frame.grid_rowconfigure(4, weight=1)
        button_constrain.grid(column=0, row=5, sticky='ews', padx=1, pady=1)
        
        self.get_input()
        
    



from tkinter import *
from tkinter import ttk, messagebox
from bfit import logger_name
from bdata import bdata, bmerged
from bfit.gui.InputLine import InputLine

import bfit.backend.colors as colors
import numpy as np
import bdata as bd

import logging, textwrap


# =========================================================================== #
class fitline(object):
    """
        Instance variables

            bfit            pointer to top class
            data            fitdata object in bfit.data dictionary
            disable_entry_callback  disables copy of entry strings to
                                    dataline.bdfit parameter values
            gui_param_button ttk.Button, set initial parameters 
            lines           list of InputLine objects
            parent          pointer to parent object (frame)
            run_label       label for showing which run is selected
            run_label_title label for showing which run is selected
            fitframe        mainframe for this tab.
    """

    collist = ['p0', 'blo', 'bhi', 'res', 'dres-', 'dres+', 'chi', 'fixed', 'shared']

    # ======================================================================= #
    def __init__(self, bfit, parent, data, row):
        """
            Inputs:
                bfit:       top level pointer
                parent:     pointer to parent frame object
                dataline:   fetch_files.dataline object corresponding to the
                                data we want to fit
                row:        grid position
        """

        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing fit line for run %d in row %d',
                          data.run, row)

        # initialize
        self.bfit = bfit
        self.parent = parent
        self.data = data
        self.row = row
        self.disable_entry_callback = False
        self.lines = []

        data.fitline = self

        # get parent frame
        fitframe = ttk.Frame(self.parent, pad=(5, 0))

        frame_title = ttk.Frame(fitframe)

        # label for displyaing run number
        if type(self.data.bd) is bdata:
            self.run_label = Label(frame_title,
                            text='[ %d - %d ]' % (self.data.run,
                                                  self.data.year),
                           bg=colors.foreground, fg=colors.background)

        elif type(self.data.bd) is bmerged:
            runs = textwrap.wrap(str(self.data.run), 5)

            self.run_label = Label(frame_title,
                                text='[ %s ]' % ' + '.join(runs),
                                bg=colors.foreground, fg=colors.background)

        # title of run
        self.run_label_title = Label(frame_title,
                            text=self.data.title,
                            justify='right', fg=colors.red)

        # Parameter input labels
        gui_param_button = ttk.Button(fitframe, text='Initial Value',
                        command=lambda : self.bfit.fit_files.do_gui_param(id=self.data.id),
                        pad=0)
        result_comp_button = ttk.Button(fitframe, text='Result',
                        command=self.draw_fn_composition, pad=0)

        c = 0
        ttk.Label(fitframe, text='Parameter').grid(     column=c, row=1, padx=5); c+=1
        gui_param_button.grid(                          column=c, row=1, padx=5, pady=2); c+=1
        ttk.Label(fitframe, text='Low Bound').grid(     column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='High Bound').grid(    column=c, row=1, padx=5); c+=1
        result_comp_button.grid(                        column=c, row=1, padx=5, pady=2, sticky=(E, W)); c+=1
        ttk.Label(fitframe, text='Error (-)').grid(     column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='Error (+)').grid(     column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='ChiSq').grid(         column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='Fixed').grid(         column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='Shared').grid(        column=c, row=1, padx=5); c+=1
        
        self.run_label.grid(column=0, row=0, padx=5, pady=5, sticky=W)
        self.run_label_title.grid(column=2, row=0, padx=5, pady=5, sticky=E)
        frame_title.grid(column=0, row=0, columnspan=c, sticky=(E, W))
        frame_title.columnconfigure(1, weight=1)

        # save frame
        self.fitframe = fitframe
        self.gui_param_button = gui_param_button

        # resizing
        for i in range(c):
            self.fitframe.grid_columnconfigure(i, weight=1)

    # ======================================================================= #
    def __del__(self):

        # kill buttons and frame
        try:
            for child in self.parent.winfo_children():
                child.destroy()
        except Exception:
            pass

        if hasattr(self, 'parentry'):    del self.parentry
        
    # ======================================================================= #
    def get_new_parameters(self, force_modify=False):
        """
            Fetch initial parameters from fitter, set to data.

            plist: Dictionary of initial parameters {par_name:par_value}
        """
        
        # get pointer to fit files object
        fit_files = self.bfit.fit_files
        fitter = fit_files.fitter
        ncomp = fit_files.n_component.get()
        fn_title = fit_files.fit_function_title.get()

        # check if we are using the fit results of the prior fit
        values_res = None
        res = self.data.fitpar['res']
        
        isfitted = not all(res.isna()) # is this run fitted?
        
        if fit_files.set_prior_p0.get() and not isfitted:
            
            r = 0
            for data in self.bfit.data.values():
                isfitted = not all(data.fitpar['res'].isna()) # is the latest run fitted?
                if isfitted and data.run > r:
                    r = data.run
                    values_res = data.fitpar.copy()
        
        # get calcuated initial values
        try:
            values = fitter.gen_init_par(fn_title, ncomp, self.data.bd,
                                    self.bfit.get_asym_mode(fit_files))
        except Exception as err:
            print(err)
            self.logger.exception(err)
            return
                
        # set p0 from old
        if values_res is not None:
            for idx in values_res.index:
                if idx not in values.index:
                    values = values.append(values_res.loc[idx])
                    values.loc[idx, ['res', 'dres+', 'dres-', 'chi']] = np.nan
            values.loc[:, 'p0'] = values_res.loc[:, 'res']
        
        # set to data
        self.data.set_fitpar(values)
        
    # ======================================================================= #
    def grid(self, row):
        """Re-grid a dataline object so that it is in order by run number"""
        self.row = row
        self.fitframe.grid(column=0, row=row, sticky=(W, N))
        self.fitframe.update_idletasks()

    # ======================================================================= #
    def degrid(self):
        """Remove displayed dataline object from file selection. """

        self.logger.debug('Degridding fitline for run %s', self.data.id)
        self.fitframe.grid_forget()
        self.fitframe.update_idletasks()

    # ======================================================================= #
    def draw_fn_composition(self):
        """
            Draw window with function components and total
        """

        self.logger.info('Drawing fit composition for run %s', self.data.id)

        # get top objects
        fit_files = self.bfit.fit_files
        bfit = self.bfit

        # get fit object
        bdfit = self.data

        # get base function
        fn_name = fit_files.fit_function_title.get()

        # get number of components and parameter names
        ncomp = fit_files.n_component.get()
        pnames_single = fit_files.fitter.gen_param_names(fn_name, 1)
        pnames_combined = fit_files.fitter.gen_param_names(fn_name, ncomp)

        if '2' in bdfit.mode:
            fn_single = fit_files.fitter.get_fn(fn_name=fn_name, 
                            ncomp=1,
                            pulse_len=bdfit.pulse_s,
                            lifetime=bd.life[bfit.probe_species.get()])
            fn_combined = fit_files.fitter.get_fn(fn_name=fn_name, 
                            ncomp=ncomp,
                            pulse_len=bdfit.pulse_s,
                            lifetime=bd.life[bfit.probe_species.get()])
        else:
            fn_single = fit_files.fitter.get_fn(fn_name=fn_name, ncomp=1)
            fn_combined = fit_files.fitter.get_fn(fn_name=fn_name, ncomp=ncomp)

        # draw in redraw mode
        draw_mode = bfit.draw_style.get()
        bfit.draw_style.set('redraw')

        # draw the data
        bdfit.draw(bfit.get_asym_mode(fit_files), figstyle='fit', color='k')

        # get the fit results
        results = {par:bdfit.fitpar.loc[par, 'res'] for par in pnames_combined}

        # draw if ncomp is 1
        if ncomp == 1:
            bfit.draw_style.set('stack')
            bdfit.draw_fit('fit', unique=False, 
                           asym_mode=bfit.get_asym_mode(self.bfit.fit_files), 
                           label=fn_name)
            self.bfit.draw_style.set(draw_mode)
            return

        # draw baseline
        if 'baseline' in pnames_single:
            bfit.plt.axhline('fit', bdfit.id+'_base', results['baseline'], ls='--', zorder=6)

        # get x pts
        t, a, da = bdfit.asym(bfit.get_asym_mode(fit_files))
        fitx = np.linspace(min(t), max(t), fit_files.n_fitx_pts)

        # get x axis scaling
        if bdfit.mode in bfit.units:
            unit = bfit.units[bdfit.mode]
        else:
            fitxx = fitx

        # draw relative to peak 0
        if self.bfit.draw_rel_peak0.get():
            
            # get reference
            par = data.fitpar
            
            if 'peak_0' in par.index:   index = 'peak_0'
            elif 'mean_0' in par.index: index = 'mean_0'
            elif 'peak' in par.index:   index = 'peak'
            elif 'mean' in par.index:   index = 'mean'
            else:
                msg = "No 'peak' or 'mean' fit parameter found. Fit with" +\
                     " an appropriate function."
                self.logger.exception(msg)
                messagebox.error(msg)
                raise RuntimeError(msg)
            
            ref = par.loc[index, 'res']
            
            # do the shift
            fitxx = fitx-ref                    
            fitxx *= unit[0]
            xlabel = 'Frequency Shift (%s)' % unit[1]
            self.logger.info('Drawing as freq shift from peak_0')
        
        # ppm shift
        elif self.bfit.draw_ppm.get():
            
            # check div zero
            try:
                fitxx = 1e6*(fitx-self.bfit.ppm_reference)/self.bfit.ppm_reference
            except ZeroDivisionError as err:
                self.logger.exception(str(msg))
                messagebox.error(str(msg))
                raise err
            
            self.logger.info('Drawing as PPM shift with reference %s Hz', 
                             self.bfit.ppm_reference)
            xlabel = 'Frequency Shift (PPM)'
            
        else: 
            fitxx = fitx*unit[0]

        # draw the combined
        params = [results[name] for name in pnames_combined]

        bfit.plt.plot('fit', bdfit.id+'_comb', fitxx, fn_combined(fitx, *params),
                            unique=False, label='Combined', zorder=5)

        # draw each component
        for i in range(ncomp):

            # get parameters
            params = [results[single+'_%d'%i] \
                        for single in pnames_single if single != 'baseline']

            if 'baseline' in pnames_single:
                params.append(results['baseline'])

            # draw
            bfit.plt.plot('fit', bdfit.id+'_%d'%i, fitxx, fn_single(fitx, *params),
                            unique=False, ls='--', label='%s %d'%(fn_name, i), zorder=6)

        # plot legend
        bfit.plt.legend('fit')

        # reset to old draw mode
        bfit.draw_style.set(draw_mode)

    # ======================================================================= #
    def populate(self, force_modify=False):
        """
            Fill and grid new parameters. Reuse old fields if possible

            force_modify: if true, clear and reset parameter inputs.
        """

        # get data and frame
        fitframe = self.fitframe
        fitdat = self.data
        fit_files = self.bfit.fit_files
        pop_constr = fit_files.pop_fitconstr

        if force_modify:
            fitdat.reset_fitpar()
            
        # check if new data
        new_data = len(fitdat.fitpar.index) == 0
        
        # disable init button if constrained
        pop_constr.set_init_button_state(self)
                    
        # add constrained parameter and function
        pop_constr.add_new_par(fitdat)    
        pop_constr.add_fn(fitdat)
        
        # drop parameters which are no longer needed
        pop_constr.drop_unused_param(fitdat)
                
        # set new p0
        if force_modify or new_data:
            
            # make a new parameter dataframe
            try:
                self.get_new_parameters(force_modify)
            except KeyError as err:
                return          # returns if no parameters found
            except RuntimeError as err:
                messagebox.showerror('RuntimeError', err)
                raise err from None
        
        # get list of parameters and initial values
        fitdat.fitpar.sort_index(inplace=True)
        fitpar = fitdat.fitpar
        plist = tuple(fitpar.index.values)
        self.logger.debug('Populating parameter list with %s', plist)
        
        # get needed number of lines
        n_lines_total = len(plist)
        n_lines_current = len(self.lines)
        n_lines_needed = n_lines_total - n_lines_current
        
        # drop unneeded lines
        if n_lines_needed < 0:
            
            unused = self.lines[n_lines_total:]
            for rem in unused:
                rem.degrid()
                del rem
            self.lines = self.lines[:n_lines_total]

        # add new lines
        elif n_lines_needed > 0:
            self.lines.extend([InputLine(fitframe, self.bfit, fitdat) \
                                                for i in range(n_lines_needed)])

        # reassign and regrid lines
        for i, line in enumerate(self.lines):
            line.grid(i+2)
            
        # drop old parameters
        fitdat.drop_unused_param(plist)
        fitdat.fitpar.sort_index(inplace=True)
        fitpar = fitdat.fitpar
        
        # set parameters
        for i, k in enumerate(fitpar.index):
            self.lines[i].set(k, **fitpar.loc[k].to_dict())
                                     
    # ======================================================================= #
    def set(self, pname, **kwargs):
        """
            Set line columns
        """
        for line in self.lines:
            if line.pname == pname:
                line.set(**kwargs)
        
    # ======================================================================= #
    def show_fit_result(self):
        self.logger.debug('Showing fit result for run %s', self.data.id)

        try:
            data = self.data
        except KeyError:
            return
        
        try:
            chi = data.chi
        except AttributeError:
            return

        # show fit results
        for line in self.lines:
            values = {r: data.fitpar.loc[line.pname, r] for r in ('res', 'dres-', 'dres+')}
            values['chi'] = chi
            line.set(**values)




# Window popup for an ongoing process
# Derek Fujimoto
# Nov 2020

from tkinter import *
from tkinter import ttk
from multiprocessing import Process
from queue import Empty

import bfit.backend.colors as colors

class popup_ongoing_process(object):
    """
        bfit:       pointer to bfit object
        do_disable: function to run to disable GUI elements on process start
        do_enable:  function to run to enable GUI elements on process end
        kill_status:BooleanVar, if true, the process was terminated
        message:    string, summary of process which is ongoing
        process:    multiprocessing.Process
        root:       TopLevels
        target:     function handle, run this with no arguments
        queue:      multiprocessing.Queue: put output in here, gets returned
    """
    
    
    def __init__(self, bfit, target, message, queue, do_disable=None, do_enable=None):
        """
            bfit:       pointer to bfit object
            do_disable: function to run to disable GUI elements on process start
            do_enable:  function to run to enable GUI elements on process end
            message:    string, summary of process which is ongoing
            root:       TopLevels
            target:     function handle, run this with no arguments
            queue:      multiprocessing.Queue: put output in here, gets returned
        """
        
        # variables
        self.bfit = bfit
        self.message = message
        self.logger = bfit.logger
        self.target = target
        self.do_disable = do_disable
        self.do_enable = do_enable
        self.queue = queue
        self.kill_status = BooleanVar()
        self.kill_status.set(False)
        
        # make window
        root = Toplevel(bfit.root)
        root.lift()
        root.resizable(FALSE, FALSE)
        
        # set icon
        self.bfit.set_icon(root)
        
        # set label
        label = ttk.Label(root, 
                      text=message, 
                      justify='center', 
                      pad=0)
        
        # make progress bar
        pbar = ttk.Progressbar(root, orient=HORIZONTAL, 
                               mode='indeterminate', length=200, maximum=20)
        pbar.start()
        
        # make button to cancel the fit
        cancel = ttk.Button(root, 
                      text="Cancel", 
                      command=self._kill,
                      pad=0)

        # grid 
        label.grid(column=0, row=0, padx=15, pady=5)
        pbar.grid(column=0, row=1, padx=15, pady=5)
        cancel.grid(column=0, row=2, padx=15, pady=5)
        
        # set up close window behaviour 
        root.protocol("WM_DELETE_WINDOW", self._kill)
        
        # update
        bfit.root.update_idletasks()
        
        # set window size
        width = root.winfo_reqwidth()
        height = root.winfo_reqheight()
        
        rt_x = bfit.root.winfo_x()
        rt_y = bfit.root.winfo_y()
        rt_w = bfit.root.winfo_width()
        rt_h = bfit.root.winfo_height()
        
        x = rt_x + rt_w/2 - (width/2)
        y = rt_y + rt_h/3 - (width/2)
        
        root.geometry('{}x{}+{}+{}'.format(width, height, int(x), int(y)))
        self.root = root
        
    def _kill(self):
        """Terminate the processs"""
        
        self.process.terminate()
        self.kill_status.set(True)
        self.logger.info('%s canceled', self.message)    
    
    def run(self):
        
        # start fit
        self.process = Process(target = self.target)
        self.process.start()
        
        # disable GUI
        if self.do_disable is not None:  
            self.do_disable()
        
        # get the output, checking for kill signal
        try:
            while True:  
                try: 
                    output = self.queue.get(timeout = 0.001)
                    
                except Empty:
                    
                    # iterate the gui
                    try:
                        self.root.update()
                    
                    # applicated destroyed
                    except TclError:    
                        return
                    
                    # check if fit cancelled
                    if self.kill_status.get():
                        if self.do_enable is not None:  
                            self.do_enable()
                        return 
                        
                # got someting in the queue
                else:
                    self.process.join()
                    if self.do_enable is not None:  
                        self.do_enable()
                    
                    # fit success
                    return output
                    
        finally:
            try:
                # kill process, destroy fit window
                self.process.terminate()
                self.root.destroy()
                del self.root
                
            # window already destroyed case (main window closed)
            except TclError:    
                pass
    
    
        
        



# Model the fit results with a function
# Derek Fujimoto
# Nov 2019


from tkinter import *
from tkinter import ttk
from functools import partial

import logging, re, os, warnings
import numpy as np
import pandas as pd
import bdata as bd
import weakref as wref

from bfit import logger_name
from bfit.backend.entry_color_set import on_focusout, on_entry_click
from bfit.backend.raise_window import raise_window
from bfit.fitting.fit_bdata import fit_bdata
from bfit.backend.ParameterFunction import ParameterFunction as ParFnGenerator
from bfit.global_variables import KEYVARS
from bfit.gui.template_fit_popup import template_fit_popup
import bfit.backend.colors as colors

# ========================================================================== #
class popup_add_param(template_fit_popup):
    """
        Popup window for adding parameters for fitting or drawing

        bfit
        fittab
        logger

        input_fn_text       string: input lines

        new_par:            dict: {parname, string: par equation}
        set_par:            dict: {parname, fn handle: lambda : return new_par}
        parnames:           list, function inputs
        reserved_pars:      dict, define values in bdata that can be accessed
        win:                Toplevel
    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}

    window_title = 'Add parameter for fitting or drawing'

    # ====================================================================== #
    def __init__(self, bfit, input_fn_text=''):


        super().__init__(bfit, input_fn_text)
        super().show()

        self.fittab = bfit.fit_files
        self.set_par = {}
        self.new_par = {}

        # add button
        add_button = ttk.Button(self.right_frame, text='Set Parameters', command=self.do_add)

        add_button.grid(column=0, row=1, sticky=(N, E, W, S), padx=20, pady=20)

        # Keyword parameters
        key_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved variable names:\n\n'
        self.reserved_pars = KEYVARS

        keys = list(self.reserved_pars.keys())
        descr = [self.reserved_pars[k] for k in self.reserved_pars]
        maxk = max(list(map(len, keys)))

        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        key_param_label = ttk.Label(key_param_frame, text=s, justify=LEFT)

        # fit parameter names
        fit_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved parameter names:\n\n'
        self.parnames = self.fittab.fitter.gen_param_names(
                                        self.fittab.fit_function_title.get(),
                                        self.fittab.n_component.get())

        s += '\n'.join([k for k in sorted(self.parnames)])
        s += '\n'
        fit_param_label = ttk.Label(fit_param_frame, text=s, justify=LEFT)

        # module names
        module_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved module names:\n\n'

        keys = list(self.modules.keys())
        descr = [self.modules[k] for k in self.modules]
        maxk = max(list(map(len, keys)))

        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        modules_label = ttk.Label(module_frame, text=s, justify=LEFT)

        # Text entry
        self.entry_label['text'] = 'Enter one parameter equation per line.'+\
                '\nLHS must use only reserved words, constants, or functions'+\
                '\nfrom the reserved modules in the parameter definition.'+\
                '\nEx: "mypar = 1/(1_T1*TEMP)"'+\
                '\n\nAccepts LaTeX input for the new parameter.'+\
                '\nEx: "$\eta_\mathrm{f}$ (Tesla) = B0 * np.exp(-amp**2)"' +\
                '\n\nValues taken as shown in fit results, with no unit scaling.'

        # gridding
        key_param_label.grid(column=0, row=0)
        fit_param_label.grid(column=0, row=0)
        modules_label.grid(column=0, row=0)

        key_param_frame.grid(column=0, row=0, rowspan=1, sticky=(E, W), padx=1, pady=1)
        module_frame.grid(column=0, row=1, sticky=(E, W), padx=1, pady=1, rowspan=2)
        fit_param_frame.grid(column=0, row=3, sticky=(E, W, N, S), padx=1, pady=1)

        self.logger.debug('Initialization success. Starting mainloop.')

    # ====================================================================== #
    def do_add(self, *_):
        """
            Add the parameter and the corresponding function
        """
        # reset draw comp
        self.bfit.fit_files.draw_components = list(self.bfit.draw_components)

        # set the parameters
        self.set_par = ''
        try:
            self.set_par = {p: ParFnGenerator(p, e, self.parnames, self.bfit) \
                        for p, e in self.new_par.items()}
        except SyntaxError: # on empty set
            pass

        self.logger.info('Added new parameters ', self.set_par)

        # update the lists
        self.bfit.fit_files.populate_param()

        # close the window
        self.cancel()

        # ====================================================================== #
    def do_after_parse(self, defined=None, eqn=None, new_par=None):

        # no input
        if defined is None:
            return

        # add result
        self.new_par = {k:e for k, e in zip(defined, eqn)}

    # ====================================================================== #
    def do_return(self, *_):
        return self.do_add()




# test inspect tab
# Derek Fujimoto
# Feb 2021
from numpy.testing import *
import numpy as np
import matplotlib.pyplot as plt
from bfit.gui.bfit import bfit
import bdata as bd
import os

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')
plt.ion()

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fileviewer
        
        try:
            function(*args, **kwargs, tab=tab, b=b)
        finally:
            b.on_closing()
            del b
            
    return wrapper

def test_fetch_20():    fetch(40123, 2020, '20')
def test_fetch_1f():    fetch(40033, 2020, '1f')
def test_fetch_1w():    fetch(40037, 2020, '1w')
def test_fetch_1n():    fetch(40011, 2020, '1n')
def test_fetch_2h():    fetch(45539, 2019, '2h')
def test_fetch_2e():    fetch(40326, 2019, '2e')
def test_draw_20():     draw(40123, 2020, '20')
def test_draw_1f():     draw(40033, 2020, '1f')
def test_draw_1w():     draw(40037, 2020, '1w')
def test_draw_1n():     draw(40011, 2020, '1n')

@pytest.mark.filterwarnings('ignore:2019')
@pytest.mark.filterwarnings('ignore:divide by zero')
@pytest.mark.filterwarnings('ignore:Warning')
@pytest.mark.filterwarnings('ignore:invalid value')
def test_draw_2h():     draw(45539, 2019, '2h')

def test_draw_2e():     draw(40326, 2019, '2e')

@with_bfit
def fetch(r, y, mode, tab=None, b=None):    
    
    tab.year.set(y)
    tab.runn.set(r)
    
    try:
        tab.get_data()
    except Exception as err: 
        print(err)
        raise AssertionError("fileviewer fetch %s (%d.%d) data" % (mode, y, r))
    
    assert_equal(tab.data.run, r, "fileviewer fetch %s (%d.%d) data accuracy" % (mode, y, r))
    
@with_bfit
def draw(r, y, mode, tab=None, b=None):

    # get data
    tab.year.set(y)
    tab.runn.set(r)
    tab.get_data()
    
    # draw
    n = len(tab.entry_asym_type['values'])
    
    for i in range(n):
        
        # switch draw types
        tab.entry_asym_type.current(i)
        draw_type = tab.asym_type.get()
        
        # draw
        try:
            tab.draw(figstyle='inspect')
        except Exception as err: 
            print(err)
            raise AssertionError("fileviewer draw %s in mode %s" % (mode, draw_type), 'inspect')
        
        b.plt.clf('inspect')
        
        if mode == '2e':
            b.do_close_all()
    
@with_bfit
def test_draw_new(tab=None, b=None):    
    
    tab.year.set(2020)
    b.draw_style.set('new')
    tab.runn.set(40123)
    
    for r in range(3):
        tab.get_data()
        tab.draw('inspect')
    assert_equal(len(b.plt.plots['inspect']), 3, 'fileviewer draw new')

@with_bfit
def test_draw_stack(tab=None, b=None):
    
    tab.year.set(2020)
    b.draw_style.set('stack')
    
    for r in (40123, 40127):
        tab.runn.set(r)
        tab.get_data()
        tab.draw('inspect')
    
    ax = b.plt.gcf('inspect').axes[0]
    assert_equal(len(ax.draw_objs), 2, 'fileviewer stack')
    
@with_bfit
def test_draw_redraw(tab=None, b=None):

    tab.year.set(2020)
    b.draw_style.set('redraw')
        
    for r in (40123, 40127):
        tab.runn.set(r)
        tab.get_data()
        tab.draw('inspect')
    
    ax = b.plt.gcf('inspect').axes[0]
    assert_equal(len(ax.draw_objs), 1, 'fileviewer redraw')
    
@with_bfit    
def test_autocomplete(tab=None, b=None):
  
    tab.year.set(2020)
    tab.runn.set(402)
    tab.get_data()
    assert_equal(tab.data.run, 40299, 'fileviewer autocomplete fetch')
    
@with_bfit    
def test_load_run_from_file(tab=None, b=None):
    
    # load a file
    dat = bd.bdata(40123, 2020)
    
    # get file location
    if 'BNMR_ARCHIVE' in os.environ:
        path = os.path.join(os.environ['BNMR_ARCHIVE'],'2020','040123.msr')
    else:
        path = os.path.join(os.environ['HOME'],'.bdata','bnmr','2020','040123.msr')
    
    # load file
    tab.load_file(filename=path)
    
    # check inputs disabled
    assert tab.entry_year['state'] == 'disabled', 'year state == disabled'
    assert tab.entry_runn['state'] == 'disabled', 'runn state == disabled'
    
    # check that run loaded
    assert tab.data.title == dat.title, 'tab data loaded correctly'
    
    # check that fetch loads correct data
    tab.get_data()
    assert tab.data.title == dat.title, 'tab data loaded correctly'
    
    # turn off load file
    tab.load_file(filename='')
  
    # check inputs enabled
    assert tab.entry_year['state'] == 'normal', 'year state == normal'
    assert tab.entry_runn['state'] == 'normal', 'runn state == normal'

    # check that runs now fetch properly as normal
    tab.runn.set(40127)
    tab.year.set(2020)
    tab.get_data()
    assert tab.data.title == bd.bdata(40127, 2020).title, 'tab data loaded correctly'
    



# Set up default fitting routines.
# Derek Fujimoto
# Aug 2018

import bfit.fitting.functions as fns
from bfit.fitting.decay_31mg import fa_31Mg
from bfit.fitting.gen_init_par import gen_init_par
from functools import partial
from collections.abc import Iterable
import numpy as np
import bdata as bd
import pandas as pd
import copy

class fitter(object):
    """
        Fitter base class for default minimizers
    """

    # needed to tell users what routine this is
    __name__ = 'base'

    # Define possible fit functions for given run modes
    function_names = {  '20':('Exp', 'Bi Exp', 'Str Exp'),
                        '2h':('Exp', 'Bi Exp', 'Str Exp'),
                        '1f':('Lorentzian', 'Gaussian', 'BiLorentzian', 'QuadLorentz', 'PseudoVoigt'),
                        '1x':('Lorentzian', 'Gaussian', 'BiLorentzian', 'QuadLorentz', 'PseudoVoigt'),
                        '1w':('Lorentzian', 'Gaussian', 'BiLorentzian', 'PseudoVoigt'),
                        '2e':('Lorentzian', 'Gaussian', 'BiLorentzian', 'QuadLorentz', 'PseudoVoigt')}

    # Define names of fit parameters:
    param_names = {     'Exp'       :('1_T1', 'amp'),
                        'Bi Exp'    :('1_T1', '1_T1b', 'fraction_b', 'amp'),
                        'Str Exp'   :('1_T1', 'beta', 'amp'),
                        'Lorentzian':('peak', 'fwhm', 'height', 'baseline'),
                        'BiLorentzian':('peak', 'fwhmA', 'heightA',
                                               'fwhmB', 'heightB', 'baseline'),
                        'QuadLorentz':('nu_0', 'nu_q', 'efgAsym', 'efgTheta', 'efgPhi',
	                               'amp0', 'amp1', 'amp2', 'amp3', 'fwhm',
                                       'baseline'),
                        'Gaussian'  :('mean', 'sigma', 'height', 'baseline'),
                        'PseudoVoigt':('peak', 'fwhm', 'height', 'fracL', 'baseline'),
                        }

    # nice parameter names for labels - must be one to one unique for later inversion
    pretty_param = {'1_T1':             r'$1 / T_1$ (s$^{-1}$)',
                    'T1':               r'$T_1$ (s)',
                    'amp':              'Amplitude',
                    'fraction_b':       r'Fraction of $T_1^{(b)}$',
                    '1_T1b':            r'$1/T_1^{(b)}$ (s$^{-1}$)',
                    'T1b':              r'$T_1^{(b)}$ (s)',
                    'beta':             r'$\beta$',
                    'fwhm':             'FWHM (%s)',
                    'height':           'Height',
                    'baseline':         'Baseline',
                    'peak':             'Peak (%s)',
                    'fwhmA':            'FWHM$_A$ (%s)',
                    'fwhmB':            'FWHM$_B$ (%s)',
                    'heightA':          'Height$_A$',
                    'heightB':          'Height$_B$',
                    'nu_0':             r'$\nu_0$ (%s)',
                    'nu_q':             r'$\nu_q$ (%s)',
                    'efgAsym':          r'$\eta$',
                    'efgTheta':         r'Principle Axes $\theta$ (rad)',
                    'efgPhi':           r'Principle Axes $\phi$ (rad)',
                    'amp0':             'Amplitude$_0$',
                    'amp1':             'Amplitude$_1$',
                    'amp2':             'Amplitude$_2$',
                    'amp3':             'Amplitude$_3$',
                    'mean':             r'$\langle\nu\rangle$ (%s)',
                    'sigma':            r'$\sigma$ (%s)',
                    'Beta-Avg 1/<T1>':  r'$1/\langle T_1\rangle_\beta$ (s$^{-1}$)',
                    'B0 Field (T)':     r'$B_0$ Field (T)',
                    'fwhmL':            r'FWHM$_{\mathrm{lorentzian}}$ (%s)',
                    'fracL':            r'Fraction Lorentzian',
                    }

    # define list of ok run modes
    valid_asym_modes = ('c', 'p', 'n', 'sl_c', 'dif_c', )

    # probe species spin numbers
    spin = {'Li8':2,    # radioactive daughters: none
            'Li9':1.5,  # radioactive daughters: Be8
            'Li11':0.5, # radioactive daughters: Be11, He6, Be10
            'Be11':0.5, # radioactive daughters: none
            'F20':2,    # radioactive daughters: none
            'Mg31':0.5, # radioactive daughters: Al31, Si31
            'Mg29':1.5, # radioactive daughters: Al29
            'Ac230':1,  # radioactive daughters: none
            'Ac232':1,  # radioactive daughters: unknown
            'Ac234':1,  # radioactive daughters: unkonwn
            }

    # ======================================================================= #
    def __init__(self, keyfn, probe_species='Li8'):
        """
            keyfn:          function takes as input bdata or bjoined or bmerged
                            object, returns string corresponding to unique id of
                            that object
            probe_species: one of the keys in the bdata.life dictionary.
        """
        self.keyfn = keyfn
        self.probe_species = probe_species

    # ======================================================================= #
    def __call__(self, fn_name, ncomp, data_list, hist_select, asym_mode, xlims):
        """
            Fitting controller.

            fn_name: name of function to fit
            ncomp : number of components to incude (2 = biexp, for example)
            data_list: list of [[fitdata object, pdict, doptions], ]

                where pdict = {par:(init val,   # initial guess
                                    bound_lo,   # lower fitting bound
                                    bound_hi,   # upper fitting bound
                                    is_fixed,   # boolean, fix value?
                                    is_shared,  # boolean, share value globally?
                                   )
                              }
                where doptions = {  'omit':str,     # bins to omit in 1F calcs
                                    'rebin':int,    # rebinning factor
                                    'group':int,    # fitting group
                                 }

            hist_select: string for selection of histograms
            asym_mode:  input for asymmetry calculation type
                                               c: combined helicity
                            h: split helicity
                            p: positive helicity
                            n: negative helicity

                        For 2e mode, prefix with:
                            sl_: combined timebins using slopes
                            dif_: combined timebins using differences
                            raw_: raw time-resolved

                            ex: sl_c or raw_h or dif_c
            xlims: fit subrange of x axis
            returns dictionary of {runid: [[par_names], [par_values], [par_errors],
                                          [chisquared], [fitfunction pointers]]}
                                   and global chisquared
        """

        # check ncomponents
        if ncomp < 1:
            raise RuntimeError('ncomp needs to be >= 1')

        asym_mode = asym_mode.replace('2e_', '')

        # parameter names
        keylist = list(self.gen_param_names(fn_name, ncomp, data_list[0][0].constrained))
        npar = len(keylist)

        # gather list of data to fit
        fn = []
        bdata_list = []
        p0 = []
        bounds = []
        omit = []
        rebin = []
        slr_bkgd_corr = []
        sharelist = np.zeros(npar, dtype=bool)
        fixedlist = []

        for data in data_list:

            # split data list into parts
            dat = data[0]
            pdict = data[1]
            doptions = data[2]

            # probe lifetime
            life = bd.life[self.probe_species]

            # get fitting function for 20 and 2h
            if dat.mode in ['20', '2h']:
                pulse = dat.pulse_s
                fn.append(self.get_fn(fn_name=fn_name,
                                      ncomp=ncomp,
                                      pulse_len=pulse,
                                      lifetime=life,
                                      constr=dat.constrained))

            # 1f functions
            else:
                fn.append(self.get_fn(fn_name=fn_name,
                                      ncomp=ncomp,
                                      pulse_len=-1,
                                      lifetime=life,
                                      constr=dat.constrained))

            # get bdata objects
            bdata_list.append(dat)

            # get initial parameters
            p0.append(tuple(pdict[k][0] for k in keylist))

            # get fitting bounds
            bound = [[], []]
            shlist = []
            fixed = []
            for k in keylist:

                # bounds
                bound[0].append(pdict[k][1])
                bound[1].append(pdict[k][2])

                # fixed
                fixed.append(pdict[k][3])

                # sharelist
                shlist.append(pdict[k][4])

            bounds.append(bound)
            fixedlist.append(fixed)

            # look for any sharelist
            sharelist += np.array(shlist)

            # rebin and omit and slr_bkgd_corr
            try:
                rebin.append(doptions['rebin'])
            except KeyError:
                rebin.append(1)

            try:
                slr_bkgd_corr.append(doptions['slr_bkgd_corr'])
            except KeyError:
                slr_bkgd_corr.append(True)

            try:
                omit.append(doptions['omit'])
            except KeyError:
                omit.append('')

        # fit data
        kwargs = {'p0':p0, 'bounds':bounds}
        pars, stds_l, stds_h, covs, chis, gchi = self._do_fit(
                                        bdata_list,
                                        fn,
                                        omit=omit,
                                        rebin=rebin,
                                        slr_bkgd_corr=slr_bkgd_corr,
                                        shared=sharelist,
                                        hist_select=hist_select,
                                        asym_mode=asym_mode,
                                        fixed=fixedlist,
                                        xlims=xlims,
                                        parnames=keylist,
                                        **kwargs)

        # set up output dataframe
        if not isinstance(chis, Iterable):   # single run
            pars = [pars]
            stds_l = [stds_l]
            stds_h = [stds_h]
            chis = [chis]

        output = {}
        for i, d in enumerate(bdata_list):
            key = self.keyfn(d)
            output[key] = pd.DataFrame({'parname':keylist,
                                        'res': pars[i],
                                        'dres+': stds_h[i],
                                        'dres-': stds_l[i],
                                        'chi': chis[i],
                                        })
            output[key].set_index('parname', inplace=True)
        return (output, gchi)

    # ======================================================================= #
    def get_fit_fn(self, fn_name, ncomp, data_list):
        """
            Get dictionary of function handles keyed by run id.
        """
        output = {}

        for data in data_list:

            # split data list into parts
            dat = data[0]
            pdict = data[1]
            doptions = data[2]

            # probe lifetime
            life = bd.life[self.probe_species]

            # get fitting function for 20 and 2h
            if dat.mode in ('20', '2h'):
                pulse = dat.pulse_s
                fn = self.get_fn(fn_name, ncomp, pulse, life, dat.constrained)

            # 1f functions
            else:
                fn = self.get_fn(fn_name, ncomp, -1, life, dat.constrained)

            # make output
            output[self.keyfn(dat)] = fn

        return output

    # ======================================================================= #
    def gen_param_names(self, fn_name, ncomp, constr=None):
        """
            Make a list of the parameter names based on the number of components.

            fn_name: name of function (should match those in param_names)
            ncomp: number of components
            constr: dict {defined (string) : [fn (unused), par names (list of str)]

            return (names)
        """

        if fn_name == '':
            return tuple()

        # get names
        names_orig = self.param_names[fn_name]

        # special case of one component
        if ncomp == 1:
            names = names_orig

        # multicomponent: make copies of everything other than the baselines
        else:
            names = []
            for c in range(ncomp):
                for n in names_orig:
                    if 'base' in n: continue
                    names.append(n+'_%d' % c)

            if 'base' in names_orig[-1]:
                names.append(names_orig[-1])

        # constrained parameter names {defined: [fn, par]}
        if constr and constr is not None:
            names = list(names)
            new_names = []
            for c in constr.keys():
                if c in names:
                    del names[names.index(c)]
                    new_names.extend(constr[c][1])
            new_names = np.unique(new_names)
            names.extend(new_names)

        return tuple(names)

    # ======================================================================= #
    def gen_init_par(self, fn_name, ncomp, bdataobj, asym_mode='combined'):
        return gen_init_par(fn_name, ncomp, bdataobj, asym_mode)

    # ======================================================================= #
    def get_fn(self, fn_name, ncomp=1, pulse_len=-1, lifetime=-1, constr=None):
        """
            Get the fitting function used.

                fn_name: string of the function name users will select.
                ncomp: number of components, ex if 2, then return exp+exp
                pulse_len: duration of beam on in s
                lifetime: lifetime of probe in s

            Returns python function(x, *pars)
        """

        # set fitting function
        if fn_name == 'Lorentzian':
            fn =  fns.lorentzian
            self.mode=1
        elif fn_name == 'BiLorentzian':
            fn =  fns.bilorentzian
            self.mode=1
        elif fn_name == 'QuadLorentz':
            fn =  lambda freq, nu_0, nu_q, eta, theta, phi, \
            		amp0, amp1, amp2, amp3, fwhm: \
                   	fns.quadlorentzian(freq, nu_0, nu_q, eta, theta, phi, \
		            		amp0, amp1, amp2, amp3, \
                		   	fwhm, fwhm, fwhm, fwhm, \
                		   	I=self.spin[self.probe_species])
            self.mode=1
        elif fn_name == 'Gaussian':
            fn =  fns.gaussian
            self.mode=1
        elif fn_name == "PseudoVoigt":
            fn =  fns.pseudo_voigt
            self.mode=1
        elif fn_name == 'Exp':
            fn =  fns.pulsed_exp(lifetime, pulse_len)
            self.mode=2
        elif fn_name == 'Bi Exp':
            fn =  fns.pulsed_biexp(lifetime, pulse_len)
            self.mode=2
        elif fn_name == 'Str Exp':
            fn =  fns.pulsed_strexp(lifetime, pulse_len)
            self.mode=2
        else:
            raise RuntimeError('Fitting function not found.')

        # add corrections for probe daughters
        if self.mode == 2 and self.probe_species == 'Mg31':
            fn = fns.decay_corrected_fn(fa_31Mg, fn, beam_pulse=pulse_len)

        # Make superimposed function based on number of components
        fnlist = [fn]*ncomp

        if self.mode == 1:
            fnlist.append(lambda x, b: b)

        fn = fns.get_fn_superpos(fnlist)

        # set parameter constraints
        if constr and constr is not None:
            par_names_orig = self.gen_param_names(fn_name, ncomp)
            par_names_constr = self.gen_param_names(fn_name, ncomp, constr)
            fn = fns.get_constrained_fn(fn, par_names_orig, par_names_constr, constr)

        return fn





# Deadtime set and calculate
# Derek Fujimoto
# Feb 2021

from tkinter import *
from tkinter import ttk, messagebox
from bfit import logger_name
import logging, webbrowser, textwrap, datetime
import bdata as bd
import numpy as np
import matplotlib.pyplot as plt

# ========================================================================== #
class popup_deadtime(object):
    """
        Popup window for finding and setting deadtime corrections. 
        
        Attributes: 
        
        bfit:       bfit object
        c:          Float, calculated helicity output
        chi:        StringVar, display chi2
        dt:         Float, calculated deadtime output
        logger:     logger 
        run:        IntVar, run number of run to fetch
        bfit.deadtime_switch:   BooleanVar, if true use calculated dt value
        win:        root Tk object
        year:       IntVar, year of run to fetch
        
    """

    # ====================================================================== #
    def __init__(self, bfit):
        self.bfit = bfit
        
        # set default deadtimes
        if bfit.deadtime_global.get():
            self.dt = bfit.deadtime*1e9
            self.c = 1
        else:
            self.dt = 0
            self.c = bfit.deadtime
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
         # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title('Find and set deadtime')
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        
        # icon
        bfit.set_icon(self.win)
        
        # Key bindings
        self.win.bind('<Return>', self.find)        
        self.win.bind('<KP_Enter>', self.find)
        self.win.bind('<Shift-Key-Return>', self.draw)        
        self.win.bind('<Shift-Key-KP_Enter>', self.draw)
        
        # Run entry ----------------------------------------------------------
        frame_run = ttk.Frame(frame, pad=5)
        
        label_year = ttk.Label(frame_run, text="Year:", pad=5, justify=LEFT)
        label_run  = ttk.Label(frame_run, text="Run:", pad=5, justify=LEFT)
        
        self.year = IntVar()
        self.run = IntVar()
        
        # set year and run
        self.year.set(self.bfit.fileviewer.year.get())
        self.run.set(self.bfit.fileviewer.runn.get())
        
        spin_year = Spinbox(frame_run, from_=2000, to=datetime.datetime.today().year, 
                            textvariable=self.year, width=7)
        spin_run = Spinbox(frame_run, from_=0, to=50000, textvariable=self.run, width=7)
        
        button_find = ttk.Button(frame_run, text='Find', command=self.find)
        button_draw = ttk.Button(frame_run, text='Draw', command=self.draw)
        
        # grid
        frame_run.grid(column=0, row=0, sticky='new', padx=2, pady=2)
        label_year.grid(column=0, row=1, sticky='nse', padx=2, pady=0)
        label_run.grid(column=0, row=2, sticky='nse', padx=2, pady=0)
        spin_year.grid(column=1, row=1, sticky='nse', padx=2, pady=2)
        spin_run.grid(column=1, row=2, sticky='nse', padx=2, pady=2)
        button_find.grid(column=3, row=1, rowspan=2, sticky='nws', padx=6, pady=2)
        button_draw.grid(column=4, row=1, rowspan=2, sticky='nws', padx=6, pady=2)
        
        frame_run.columnconfigure(2, weight=1)
        
        # input ---------------------------------------------------------------
        frame_input = ttk.Frame(frame, pad=5)
        
        label_dt = ttk.Label(frame_input, text='Deadtime (ns)', pad=2, justify=LEFT)
        label_c = ttk.Label(frame_input, text='Helicity Scaling (-)', pad=2, justify=LEFT)
        
        self.dt_inpt = StringVar()
        self.c_inpt = StringVar()
        
        self.dt_inpt.set('%g' % self.dt)
        self.c_inpt.set('%g' % self.c)
        
        entry_dt = Entry(frame_input, textvariable=self.dt_inpt, width=10, justify=RIGHT)
        entry_c =  Entry(frame_input, textvariable=self.c_inpt,  width=10, justify=RIGHT)
        
        entry_dt.bind('<KeyRelease>', self.read_dt)
        entry_c.bind('<KeyRelease>',  self.read_c)
        
        self.fix_dt = BooleanVar()
        self.fix_c = BooleanVar()
        
        self.fix_dt.set(False)
        self.fix_c.set(True)
        
        check_fix_dt = ttk.Checkbutton(frame_input, text='Fix', pad=5,
                            variable=self.fix_dt, onvalue=True, offvalue=False) 
        check_fix_c = ttk.Checkbutton(frame_input, text='Fix', pad=5,
                            variable=self.fix_c, onvalue=True, offvalue=False) 
        
        # gridding
        frame_input.grid(column=0, row=1, sticky='new', padx=2, pady=2)
        
        label_dt.grid(column=0, row=0, sticky='sew', padx=2)
        entry_dt.grid(column=1, row=0, sticky='sew', padx=2)
        check_fix_dt.grid(column=2, row=0, sticky='sew', padx=2)
        
        label_c.grid(column=0, row=1, sticky='sew', padx=2)
        entry_c.grid(column=1, row=1, sticky='sew', padx=2)
        check_fix_c.grid(column=2, row=1, sticky='sew', padx=2)
        
        # chi2 ----------------------------------------------------------------
        self.chi = StringVar()
        self.chi.set('')
        label_chi = ttk.Label(frame, textvariable=self.chi, pad=2, justify=LEFT)
        label_chi.grid(column=0, row=2, sticky='new', padx=2, pady=2)
        
        # global/local switch -------------------------------------------------
        self.check_global = ttk.Checkbutton(frame, 
                text='Using deadtime of %.3f ns\nglobally' % self.dt, 
                variable=self.bfit.deadtime_global, onvalue=True, offvalue=False, 
                pad=5, command=self.toggle_scope)
        self.check_global.grid(column=0, row=3, sticky='new', padx=2, pady=2)
            
        # apply correction ---------------------------------------------------
        self.check_corr = ttk.Checkbutton(frame, 
                text='Activate deadtime correction', 
                variable=self.bfit.deadtime_switch, onvalue=True, offvalue=False, 
                pad=5)
        self.check_corr.grid(column=0, row=4, sticky='new', padx=2, pady=0)
            
        # grid frames --------------------------------------------------------
        frame.grid(column=0, row=0)
        self.logger.debug('Initialization success. Starting mainloop.')
        
    # ====================================================================== #
    def draw(self, *args):
        """
            Draw deadtime corrected data
        """
        
        # get data
        try:
            data = bd.bdata(self.run.get(), self.year.get())
        except bd.InputError as msg:
            messagebox.showerror('Bad run input', str(msg))
            raise msg
        
        asym = data.asym('hel')
        asym_dt = data.asym('hel', deadtime=self.dt*1e-9)
        
        asym_dt['n'] = list(asym_dt['n'])
        asym_dt['n'][0] *= self.c
        asym_dt['n'][1] *= self.c
        
        
        # draw split helicity ------------------------------------------------
        plt.figure()
        plt.errorbar(asym['time_s'], *asym['p'], fmt='.C0', zorder=0, 
                     label='Uncorrected')
        plt.errorbar(asym['time_s'], *asym['n'], fmt='.C0', zorder=0)
        plt.errorbar(asym_dt['time_s'], *asym_dt['p'], fmt='.C3', zorder=5, alpha=0.4, 
                     label='Corrected')
        plt.errorbar(asym_dt['time_s'], *asym_dt['n'], fmt='.C3', zorder=5, alpha=0.4 )
        
        # plot elements
        plt.ylabel('Asymmetry')
        plt.xlabel('Time (s)')
        plt.title('Run %d.%d\nDeadtime correction of %.3f ns' % \
            (self.year.get(), self.run.get(), self.dt) +\
            '\nNeg Helicity Scaling of %.3f' % self.c,
            fontsize='x-small')
        plt.legend(fontsize='small')
        plt.tight_layout()
        
        # draw helicity difference -------------------------------------------
        dasym = 0.5*(asym['p'][0] + asym['n'][0])
        ddasym = 0.5*(asym['p'][1]**2 + asym['n'][1]**2)**0.5
        
        dasym_sub = dasym - np.mean(dasym)
        ddasym_sub = (dasym**2 + np.std(dasym)**2/len(dasym))**2
        
        dasym_dt = 0.5*(asym_dt['p'][0] + asym_dt['n'][0])
        ddasym_dt = 0.5*(asym_dt['p'][1]**2 + asym_dt['n'][1]**2)**0.5
        
        dasym_dt_sub = dasym_dt - np.mean(dasym_dt)
        ddasym_dt_sub = (dasym_dt**2 + np.std(dasym_dt)**2/len(dasym_dt))**2
        
        plt.figure()
        
        plt.errorbar(asym['time_s'], dasym_sub, ddasym_sub, fmt='.C0', zorder=0, 
                     label='Uncorrected')
        plt.errorbar(asym_dt['time_s'], dasym_dt_sub, ddasym_dt_sub, fmt='.C3', 
                     zorder=5, alpha=0.4, label='Corrected')
        
        plt.axhline(0, ls='-', color='k', zorder=10)
        
        # plot elements
        plt.ylabel(r'$\frac{1}{2}(\mathcal{A}_+ + c\mathcal{A}_-) - ' +\
                     r'\frac{1}{2} \overline{(\mathcal{A}_+ + c\mathcal{A}_-)}$', 
                     fontsize='small')
        plt.xlabel('Time (s)')
        plt.title('Run %d.%d\nDeadtime correction of %.3f ns' % \
            (self.year.get(), self.run.get(), self.dt) +\
            '\nNeg Helicity Scaling: c = %.3f' % self.c,
            fontsize='x-small')
        plt.legend(fontsize='small')
        plt.tight_layout()
        
    # ====================================================================== #
    def find(self, *args):
        """
            Find deadtime of entered run
        """
        
        # get data
        try:
            data = bd.bdata(self.run.get(), self.year.get())
        except RuntimeError as msg:
            messagebox.showerror('Bad run input', str(msg))
            raise msg
        
        # get fix state
        fixed = []
        if self.fix_dt.get():   fixed.append('dt')
        if self.fix_c.get():    fixed.append('c')
        
        # get initial values
        p0 = {'dt': self.dt*1e-9,
              'c':  self.c}
        
        # find the correction
        try:
            m = data.get_deadtime(**p0, fixed=fixed, return_minuit=True)
        except RuntimeError as msg:
            messagebox.showerror('Bad run input', str(msg))
            raise msg
        except bd.exceptions.MinimizationError as msg:
            messagebox.showerror('Minimization failed',str(msg))
            raise msg
        
        self.dt = m.values['dt_ns']
        self.c = m.values['c']
        chi2 = m.fval
        
        # set the strings
        self.dt_inpt.set('%f' % self.dt)
        self.c_inpt.set('%f' % self.c)
        self.chi.set('Flattening χ2 = %.3f' % chi2)
        
        # set the value
        self.toggle_scope()    
        
        # activate deadtime usage
        self.bfit.deadtime_switch.set(True)
        
    # ====================================================================== #
    def read_c(self, *args):
        """
            read float scaling factor values from text input
        """
        
        try:
            self.c = float(self.c_inpt.get())
        except ValueError:
            return
        
        self.toggle_scope()
        
    # ====================================================================== #
    def read_dt(self, *args):
        """
            read float deadtime values from text input
        """
        
        try:
            self.dt = float(self.dt_inpt.get())
        except ValueError:
            return
        
        self.toggle_scope()
    
    # ====================================================================== #
    def set_deadtime(self, *args):
        """
            Set the bfit deadtime value depending on the state of 
            bfit.deadtime_global
        """
        
        # set deadtime value for globally applied deadtime
        if self.bfit.deadtime_global.get():
            self.bfit.deadtime = self.dt*1e-9
        
        # set scaling factor for locally calculated deadtimes
        else:
            self.bfit.deadtime = self.c
            
    # ====================================================================== #
    def toggle_scope(self, *args):
        """
            Change the deadtime from globally applied to locally calculated for 
            each run
        """
        
        self.set_deadtime()
        
        if self.bfit.deadtime_global.get():            
            outstring = 'Using deadtime of %.3f ns\nglobally' % (self.bfit.deadtime*1e9)
            self.logger.info('Set bfit.deadtime to global %f s', self.bfit.deadtime)
        else:
            outstring = 'Using scaling of %.3f to find\ndeadtimes for each run' % self.bfit.deadtime    
            self.logger.info('Set bfit.deadtime to %f for local deadtime calculations', \
                             self.bfit.deadtime)
        
        self.check_global.config(text=outstring)
        



# fit_files tab for bfit
# Derek Fujimoto
# Dec 2017

from tkinter import *
from tkinter import ttk, messagebox, filedialog
from bdata import bdata, bmerged
from bfit import logger_name, __version__
from pandas.plotting import register_matplotlib_converters
from multiprocessing import Queue

from bfit.gui.popup_show_param import popup_show_param
from bfit.gui.popup_param import popup_param
from bfit.gui.popup_fit_results import popup_fit_results
from bfit.gui.popup_fit_constraints import popup_fit_constraints
from bfit.gui.popup_add_param import popup_add_param
from bfit.gui.popup_ongoing_process import popup_ongoing_process
from bfit.backend.entry_color_set import on_focusout, on_entry_click
from bfit.backend.raise_window import raise_window
from bfit.gui.fitline import fitline

import numpy as np
import pandas as pd
import bdata as bd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import bfit.backend.colors as colors

import datetime, os, traceback, warnings, logging

register_matplotlib_converters()

# =========================================================================== #
class fit_files(object):
    """
        Data fields:
            annotation:     stringvar: name of quantity for annotating parameters
            annotation_combobox: box for choosing annotation label parameter
            asym_type:      asymmetry calculation type
            canvas_frame_id:id number of frame in canvas
            chi_threshold:  if chi > thres, set color to red
            draw_components:list of titles for labels, options to export, draw.
            entry_asym_type:combobox for asym calculations
            fit_canvas:     canvas object allowing for scrolling
            fit_data_tab:   containing frame (for destruction)
            fit_function_title: StringVar, title of fit function to use
            fit_function_title_box: combobox for fit function names
            fit_input:      fitting input values = (fn_name, ncomp, data_list)
            fit_lines:      Dict storing fitline objects
            fit_lines_old: dictionary of previously used fitline objects, keyed by run
            fit_routine_label: label for fit routine
            fitter:         fitting object from self.bfit.routine_mod
            gchi_label:     Label for global chisquared
            mode:           what type of run is this.

            n_component:    number of fitting components (IntVar)
            n_component_box:Spinbox for number of fitting components
            par_label       StringVar, label for plotting parameter set
            par_label_entry:draw parameter label entry box
            plt:            self.bfit.plt

            pop_addpar:     popup for ading parameters which are combinations of others
            pop_fitres:     modelling popup, for continuity between button presses
            pop_fitconstr:  object for fitting with constrained functions

            probe_label:    Label for probe species
            runframe:       frame for displaying fit results and inputs
            runmode_label:  display run mode
            set_as_group:   BooleanVar() if true, set fit parfor whole group
            set_prior_p0:   BooleanVar() if true, set P0 of newly added runs to
                            P0 of fit with largest run number
            share_var:      BooleanVar() holds share checkbox for all fitlines
            use_rebin:      BoolVar() for rebinning on fitting
            xaxis:          StringVar() for parameter to draw on x axis
            yaxis:          StringVar() for parameter to draw on y axis
            xaxis_combobox: box for choosing x axis draw parameter
            yaxis_combobox: box for choosing y axis draw parameter

            xlo, hi:         StringVar, fit range limits on x axis
    """

    mode = ""
    chi_threshold = 1.5 # threshold for red highlight on bad fits
    n_fitx_pts = 500    # number of points to draw in fitted curves

    # ======================================================================= #
    def __init__(self, fit_data_tab, bfit):

        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing')

        # initialize
        self.bfit = bfit
        self.fit_output = {}
        self.share_var = {}
        self.fitter = self.bfit.routine_mod.fitter(keyfn = bfit.get_run_key,
                                                   probe_species = bfit.probe_species.get())
        self.draw_components = list(bfit.draw_components)
        self.fit_data_tab = fit_data_tab
        self.plt = self.bfit.plt

        self.fit_lines = {}
        self.fit_lines_old = {}

        self.pop_fitconstr = popup_fit_constraints(self.bfit, self)

        # additional button bindings
        self.bfit.root.bind('<Control-Key-u>', self.update_param)

        # make top level frames
        mid_fit_frame = ttk.Labelframe(fit_data_tab,
                                       text='Initial Parameters and Fit Results', pad=5)

        mid_fit_frame.grid(column=0, row=1, rowspan=6, sticky=(S, W, E, N), padx=5, pady=5)

        fit_data_tab.grid_columnconfigure(0, weight=1)   # fitting space
        fit_data_tab.grid_rowconfigure(6, weight=1)      # push bottom window in right frame to top
        mid_fit_frame.grid_columnconfigure(0, weight=1)
        mid_fit_frame.grid_rowconfigure(0, weight=1)

        # TOP FRAME -----------------------------------------------------------

        # fit function select
        fn_select_frame = ttk.Labelframe(fit_data_tab, text='Fit Function')
        self.fit_function_title = StringVar()
        self.fit_function_title.set("")
        self.fit_function_title_box = ttk.Combobox(fn_select_frame,
                textvariable=self.fit_function_title, state='readonly')
        self.fit_function_title_box.bind('<<ComboboxSelected>>',
                lambda x:self.populate_param(force_modify=True))

        # number of components in fit spinbox
        self.n_component = IntVar()
        self.n_component.set(1)
        self.n_component_box = Spinbox(fn_select_frame, from_=1, to=20,
                textvariable=self.n_component, width=5,
                command=lambda:self.populate_param(force_modify=True))

        # fit and other buttons
        fit_button = ttk.Button(fn_select_frame, text='        Fit        ', command=self.do_fit, \
                                pad=1)
        constraint_button = ttk.Button(fn_select_frame, text='Constrain',
                                       command=self.show_constr_window, pad=1)
        set_param_button = ttk.Button(fn_select_frame, text='   Set Result as P0   ',
                        command=self.do_set_result_as_initial, pad=1)
        reset_param_button = ttk.Button(fn_select_frame, text='     Reset P0     ',
                        command=self.do_reset_initial, pad=1)

        # GRIDDING

        # top frame gridding
        fn_select_frame.grid(column=0, row=0, sticky=(W, E, N), padx=5, pady=5)

        c = 0
        self.fit_function_title_box.grid(column=c, row=0, padx=5); c+=1
        ttk.Label(fn_select_frame, text="Number of Terms:").grid(column=c,
                  row=0, padx=5, pady=5, sticky=W); c+=1
        self.n_component_box.grid(column=c, row=0, padx=5, pady=5, sticky=W); c+=1
        fit_button.grid(column=c, row=0, padx=5, pady=1, sticky=W); c+=1
        constraint_button.grid(column=c, row=0, padx=5, pady=1, sticky=(W, E)); c+=1
        set_param_button.grid(column=c, row=0, padx=5, pady=1, sticky=W); c+=1
        reset_param_button.grid(column=c, row=0, padx=5, pady=1, sticky=W); c+=1

        # MID FRAME -----------------------------------------------------------

        # Scrolling frame to hold fitlines
        yscrollbar = ttk.Scrollbar(mid_fit_frame, orient=VERTICAL)
        self.fit_canvas = Canvas(mid_fit_frame, bd=0,                # make a canvas for scrolling
                yscrollcommand=yscrollbar.set,                      # scroll command receive
                scrollregion=(0, 0, 5000, 5000), confine=True)       # default size
        yscrollbar.config(command=self.fit_canvas.yview)            # scroll command send
        self.runframe = ttk.Frame(self.fit_canvas, pad=5)           # holds

        self.canvas_frame_id = self.fit_canvas.create_window((0, 0),    # make window which can scroll
                window=self.runframe,
                anchor='nw')
        self.runframe.bind("<Configure>", self.config_canvas) # bind resize to alter scrollable region
        self.fit_canvas.bind("<Configure>", self.config_runframe) # bind resize to change size of contained frame

        # gridding
        self.fit_canvas.grid(column=0, row=0, sticky=(E, W, S, N))
        yscrollbar.grid(column=1, row=0, sticky=(W, S, N))

        self.runframe.grid_columnconfigure(0, weight=1)
        self.fit_canvas.grid_columnconfigure(0, weight=1)
        self.fit_canvas.grid_rowconfigure(0, weight=1)

        self.runframe.bind("<Configure>", self.config_canvas) # bind resize to alter scrollable region
        self.fit_canvas.bind("<Configure>", self.config_runframe) # bind resize to change size of contained frame

        # RIGHT FRAME ---------------------------------------------------------

        # run mode
        fit_runmode_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5),
                text='Run Mode', )
        self.fit_runmode_label = ttk.Label(fit_runmode_label_frame, text="", justify=CENTER)

        # fitting routine
        fit_routine_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5),
                text='Minimizer', )
        self.fit_routine_label = ttk.Label(fit_routine_label_frame, text="",
                                           justify=CENTER)

        # probe species
        probe_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5),
                text='Probe', )
        self.probe_label = ttk.Label(probe_label_frame,
                                     text=self.bfit.probe_species.get(),
                                     justify=CENTER)

        # global chisquared
        gchi_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5),
                text='Global ChiSq', )
        self.gchi_label = ttk.Label(gchi_label_frame, text='', justify=CENTER)

        # asymmetry calculation
        asym_label_frame = ttk.Labelframe(fit_data_tab, pad=(5, 5, 5, 5),
                text='Asymmetry Calculation', )
        self.asym_type = StringVar()
        self.asym_type.set('')
        self.entry_asym_type = ttk.Combobox(asym_label_frame, \
                textvariable=self.asym_type, state='readonly', \
                width=20)
        self.entry_asym_type['values'] = ()

        # other settings
        other_settings_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5),
                text='Switches', )

        # set as group checkbox
        self.set_as_group = BooleanVar()
        set_group_check = ttk.Checkbutton(other_settings_label_frame,
                text='Modify for all', \
                variable=self.set_as_group, onvalue=True, offvalue=False)
        self.set_as_group.set(False)

        # rebin checkbox
        self.use_rebin = BooleanVar()
        set_use_rebin = ttk.Checkbutton(other_settings_label_frame,
                text='Rebin data (set in fetch)', \
                variable=self.use_rebin, onvalue=True, offvalue=False)
        self.use_rebin.set(False)

        # set P0 as prior checkbox
        self.set_prior_p0 = BooleanVar()
        set_prior_p0 = ttk.Checkbutton(other_settings_label_frame,
                text='Set P0 of new run to prior result', \
                variable=self.set_prior_p0, onvalue=True, offvalue=False)
        self.set_prior_p0.set(False)

        # specify x axis --------------------
        xspecify_frame = ttk.Labelframe(fit_data_tab,
            text='Restrict x limits', pad=5)

        self.xlo = StringVar()
        self.xhi = StringVar()
        self.xlo.set('-inf')
        self.xhi.set('inf')

        entry_xspecify_lo = Entry(xspecify_frame, textvariable=self.xlo, width=10)
        entry_xspecify_hi = Entry(xspecify_frame, textvariable=self.xhi, width=10)
        label_xspecify = ttk.Label(xspecify_frame, text=" < x < ")

        # fit results -----------------------
        results_frame = ttk.Labelframe(fit_data_tab,
            text='Fit Results and Run Conditions', pad=5)     # draw fit results

        # draw and export buttons
        button_frame = Frame(results_frame)
        draw_button = ttk.Button(button_frame, text='Draw', command=self.draw_param)
        update_button = ttk.Button(button_frame, text='Update', command=self.update_param)
        export_button = ttk.Button(button_frame, text='Export', command=self.export)
        show_button = ttk.Button(button_frame, text='Compare', command=self.show_all_results)
        model_fit_button = ttk.Button(button_frame, text='Fit a\nModel',
                                      command=self.do_fit_model)

        # menus for x and y values
        x_button = ttk.Button(results_frame, text="x axis:", command=self.do_add_param, pad=0)
        y_button = ttk.Button(results_frame, text="y axis:", command=self.do_add_param, pad=0)
        ann_button = ttk.Button(results_frame, text=" Annotation:", command=self.do_add_param, pad=0)
        label_label = ttk.Label(results_frame, text="Label:")

        self.xaxis = StringVar()
        self.yaxis = StringVar()
        self.annotation = StringVar()
        self.par_label = StringVar()

        self.xaxis.set('')
        self.yaxis.set('')
        self.annotation.set('')
        self.par_label.set('')

        self.xaxis_combobox = ttk.Combobox(results_frame, textvariable=self.xaxis,
                                      state='readonly', width=19)
        self.yaxis_combobox = ttk.Combobox(results_frame, textvariable=self.yaxis,
                                      state='readonly', width=19)
        self.annotation_combobox = ttk.Combobox(results_frame,
                                      textvariable=self.annotation,
                                      state='readonly', width=19)
        self.par_label_entry = Entry(results_frame,
                                    textvariable=self.par_label, width=21)

        # gridding
        button_frame.grid(column=0, row=0, columnspan=2)
        draw_button.grid(column=0, row=0, padx=5, pady=5)
        update_button.grid(column=0, row=1, padx=5, pady=5)
        show_button.grid(column=1, row=0, padx=5, pady=5)
        export_button.grid(column=1, row=1, padx=5, pady=5)
        model_fit_button.grid(column=2, row=0, rowspan=2, pady=5, sticky=(N, S))

        x_button.grid(column=0, row=1, sticky=(E, W), padx=5)
        y_button.grid(column=0, row=2, sticky=(E, W), padx=5)
        ann_button.grid(column=0, row=3, sticky=(E, W), padx=5)
        label_label.grid(column=0, row=4, sticky=(E, W), padx=10)

        self.xaxis_combobox.grid(column=1, row=1, pady=5)
        self.yaxis_combobox.grid(column=1, row=2, pady=5)
        self.annotation_combobox.grid(column=1, row=3, pady=5)
        self.par_label_entry.grid(column=1, row=4, pady=5)

        # save/load state -----------------------
        state_frame = ttk.Labelframe(fit_data_tab, text='Program State', pad=5)
        state_save_button = ttk.Button(state_frame, text='Save', command=self.bfit.save_state)
        state_load_button = ttk.Button(state_frame, text='Load', command=self.bfit.load_state)

        state_save_button.grid(column=1, row=0, padx=5, pady=5)
        state_load_button.grid(column=2, row=0, padx=5, pady=5)
        state_frame.columnconfigure([0, 3], weight=1)

        # gridding
        fit_runmode_label_frame.grid(column=1, row=0, pady=5, padx=2, sticky=(N, E, W))
        self.fit_runmode_label.grid(column=0, row=0, sticky=(E, W))

        fit_routine_label_frame.grid(column=2, row=0, pady=5, padx=2, sticky=(N, E, W))
        self.fit_routine_label.grid(column=0, row=0, sticky=(E, W))

        probe_label_frame.grid(column=1, row=1, columnspan=1, sticky=(E, W, N), pady=2, padx=2)
        self.probe_label.grid(column=0, row=0)

        gchi_label_frame.grid(column=2, row=1, columnspan=1, sticky=(E, W, N), pady=2, padx=2)
        self.gchi_label.grid(column=0, row=0)

        asym_label_frame.grid(column=1, row=2, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        asym_label_frame.columnconfigure([0, 2], weight=1)
        self.entry_asym_type.grid(column=1, row=0)

        other_settings_label_frame.grid(column=1, row=3, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        set_group_check.grid(column=0, row=0, padx=5, pady=1, sticky=W)
        set_use_rebin.grid(column=0, row=1, padx=5, pady=1, sticky=W)
        set_prior_p0.grid(column=0, row=2, padx=5, pady=1, sticky=W)

        entry_xspecify_lo.grid(column=1, row=0)
        label_xspecify.grid(column=2, row=0)
        entry_xspecify_hi.grid(column=3, row=0)
        xspecify_frame.columnconfigure([0, 4], weight=1)

        xspecify_frame.grid(column=1, row=4, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        results_frame.grid(column=1, row=5, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        state_frame.grid(column=1, row=6, columnspan=2, sticky=(E, W, N), pady=2, padx=2)

        # resizing

        # fn select
        fn_select_frame.grid_columnconfigure(1, weight=1)    # Nterms label
        fn_select_frame.grid_columnconfigure(4, weight=100)    # constraints
        fn_select_frame.grid_columnconfigure(5, weight=1)  # set results as p0
        fn_select_frame.grid_columnconfigure(6, weight=1)  # reset p0

        # fitting frame
        self.fit_canvas.grid_columnconfigure(0, weight=1)    # fetch frame
        self.fit_canvas.grid_rowconfigure(0, weight=1)

        # right frame
        for i in range(2):
            results_frame.grid_columnconfigure(i, weight=0)

    # ======================================================================= #
    def __del__(self):

        if hasattr(self, 'fit_lines'):       del self.fit_lines
        if hasattr(self, 'fit_lines_old'):   del self.fit_lines_old
        if hasattr(self, 'fitter'):          del self.fitter

        # kill buttons and frame
        try:
            for child in self.fetch_data_tab.winfo_children():
                child.destroy()
            self.fetch_data_tab.destroy()
        except Exception:
            pass

     # ======================================================================= #
    def _annotate(self, id, x, y, ptlabels, color='k', unique=True):
        """Add annotation"""

        # base case
        if ptlabels is None: return

        # do annotation
        for label, xcoord, ycoord in zip(ptlabels, x, y):
            if type(label) != type(None):
                self.plt.annotate('param', id, label,
                             xy=(xcoord, ycoord),
                             xytext=(-3, 20),
                             textcoords='offset points',
                             ha='right',
                             va='bottom',
                             bbox=dict(boxstyle='round, pad=0.1',
                                       fc=color,
                                       alpha=0.1),
                             arrowprops=dict(arrowstyle = '->',
                                             connectionstyle='arc3, rad=0'),
                             fontsize='xx-small',
                             unique=unique
                            )

    # ======================================================================= #
    def canvas_scroll(self, event):
        """Scroll canvas with files selected."""
        if event.num == 4:
            self.fit_canvas.yview_scroll(-1, "units")
        elif event.num == 5:
            self.fit_canvas.yview_scroll(1, "units")

    # ======================================================================= #
    def config_canvas(self, event):
        """Alter scrollable region based on canvas bounding box size.
        (changes scrollbar properties)"""
        self.fit_canvas.configure(scrollregion=self.fit_canvas.bbox("all"))

    # ======================================================================= #
    def config_runframe(self, event):
        """Alter size of contained frame in canvas. Allows for inside window to
        be resized with mouse drag"""
        self.fit_canvas.itemconfig(self.canvas_frame_id, width=event.width)

    # ======================================================================= #
    def populate(self, *args):
        """
            Make tabs for setting fit input parameters.
        """

        # get data
        dl = self.bfit.fetch_files.data_lines
        keylist = [k for k in dl.keys() if dl[k].check_state.get()]
        keylist.sort()
        self.logger.debug('Populating data for %s', keylist)

        # get run mode by looking at one of the data dictionary keys
        for key_zero in self.bfit.data.keys(): break

        # create fit function combobox options
        try:
            if self.mode != self.bfit.data[key_zero].mode:

                # set run mode
                self.mode = self.bfit.data[key_zero].mode
                self.fit_runmode_label['text'] = \
                        self.bfit.fetch_files.runmode_relabel[self.mode]
                self.logger.debug('Set new run mode %s', self.mode)

                # set routine
                self.fit_routine_label['text'] = self.fitter.__name__

                # set run functions
                fn_titles = self.fitter.function_names[self.mode]
                self.fit_function_title_box['values'] = fn_titles

                # set current function
                if self.fit_function_title.get() not in fn_titles:
                    self.fit_function_title.set(fn_titles[0])
                    self.n_component.set(1)

        except UnboundLocalError:
            self.fit_function_title_box['values'] = ()
            self.fit_function_title.set("")
            self.fit_runmode_label['text'] = ""
            self.mode = ""
            self.n_component.set(1)

        # delete unused fitline objects
        for k in list(self.fit_lines.keys()):       # iterate fit list
            self.fit_lines[k].degrid()
            if k not in keylist:                    # check data list
                self.fit_lines_old[k] = self.fit_lines[k]
                del self.fit_lines[k]

        # make or regrid fitline objects
        n = 0
        parlst = list(self.fitter.gen_param_names(self.fit_function_title.get(),
                                                  self.n_component.get()))
        for k in keylist:
            if k not in self.fit_lines.keys():

                # add back old fit line
                if k in self.fit_lines_old.keys():
                    self.fit_lines[k] = self.fit_lines_old[k]
                    del self.fit_lines_old[k]

                # make new fit line
                else:
                    self.fit_lines[k] = fitline(self.bfit, self.runframe, dl[k].bdfit, n)

            self.fit_lines[k].grid(n)
            n+=1

        self.populate_param()

    # ======================================================================= #
    def populate_param(self, *args, force_modify=False):
        """
            Populate the list of parameters

            force_modify: passed to line.populate
        """

        self.logger.debug('Populating fit parameters')

        # populate axis comboboxes
        lst = self.draw_components.copy()

        # get list of fit parameters
        try:
            parlst = list(self.fitter.gen_param_names(
                                            self.fit_function_title.get(),
                                            self.n_component.get()),
                          )
        except KeyError:
            self.xaxis_combobox['values'] = []
            self.yaxis_combobox['values'] = []
            self.annotation_combobox['values'] = []
            return

        # set contraints flag
        constr_pop = self.pop_fitconstr
        if force_modify:
            constr_pop.constraints_are_set = False

        # get constrained parameters
        else:
            for d, new in zip(constr_pop.defined, constr_pop.new_par):
                if d in parlst:
                    parlst.extend(new)

        # Sort the parameters
        parlst = list(np.unique(parlst))
        parlst.sort()

        # add parameter beta averaged T1
        if self.fit_function_title.get() == 'Str Exp':
            ncomp = self.n_component.get()

            if ncomp > 1:
                for i in range(ncomp):
                    parlst.append('Beta-Avg 1/<T1>_%d' % i)
            else:
                parlst.append('Beta-Avg 1/<T1>')

        # add parameter T1 not 1/T1
        T1_lst = []
        if 'Exp' in self.fit_function_title.get():
            ncomp = self.n_component.get()

            if ncomp > 1:
                for i in range(ncomp):
                    T1_lst.append('T1_%d' % i)

                    if self.fit_function_title.get() == 'Bi Exp':
                        T1_lst.append('T1b_%d' % i)

            else:
                T1_lst.append('T1')

                if self.fit_function_title.get() == 'Bi Exp':
                    T1_lst.append('T1b')

        if self.fit_function_title.get() == 'Str Exp':
            if ncomp > 1:
                for i in range(ncomp):
                    T1_lst.append('Beta-Avg <T1>_%d' % i)
            else:
                T1_lst.append('Beta-Avg <T1>')

        self.xaxis_combobox['values'] = [''] + parlst + ['all the above'] + T1_lst + lst
        self.yaxis_combobox['values'] = [''] + parlst + ['all the above'] + T1_lst + lst
        self.annotation_combobox['values'] = [''] + parlst + lst

        # turn off modify all so we don't cause an infinite loop
        modify_all_value = self.set_as_group.get()
        self.set_as_group.set(False)

        # regenerate fitlines
        for fline in self.fit_lines.values():
            fline.populate(force_modify=force_modify)

        for fline in self.fit_lines_old.values():
            if self.mode == fline.data.mode:
                fline.populate(force_modify=force_modify)

        # reset modify all value
        self.set_as_group.set(modify_all_value)

    # ======================================================================= #
    def do_add_param(self, *args):
        """Launch popup for adding user-defined parameters to draw"""

        self.logger.info('Launching add paraemeter popup')

        if hasattr(self, 'pop_addpar'):
            p = self.pop_addpar

            # don't make more than one window
            if Toplevel.winfo_exists(p.win):
                p.win.lift()
                return

            # make a new window, using old inputs and outputs
            self.pop_addpar = popup_add_param(self.bfit,
                                    input_fn_text=p.input_fn_text)

        # make entirely new window
        else:
            self.pop_addpar = popup_add_param(self.bfit)

    # ======================================================================= #
    def do_end_of_fit(self):
        """Things to do after fitting: draw, set checkbox status"""

        # enable fit checkboxes on fetch files tab
        for k in self.bfit.fetch_files.data_lines.keys():
            dline = self.bfit.fetch_files.data_lines[k]
            dline.draw_fit_checkbox['state'] = 'normal'
            dline.draw_res_checkbox['state'] = 'normal'
            dline.check_fit.set(True)
        self.bfit.fetch_files.check_state_fit.set(True)

        # change fetch asymmetry mode to match fit tab
        inv_map = {v: k for k, v in self.bfit.asym_dict.items()}
        asym_mode_fit = inv_map[self.bfit.get_asym_mode(self)]
        asym_mode_fetch = inv_map[self.bfit.get_asym_mode(self.bfit.fetch_files)]

        self.bfit.fetch_files.asym_type.set(asym_mode_fit)

        # draw fit results
        if self.bfit.draw_fit.get():
            style = self.bfit.draw_style.get()

            if style in ['stack', 'new']:
                self.bfit.draw_style.set('redraw')

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self.bfit.fetch_files.draw_all(figstyle='fit', ignore_check=False)

            if len(self.fit_lines.keys()) > self.bfit.legend_max_draw:

                try:
                    self.plt.gca('fit').get_legend().remove()
                except AttributeError:
                    pass
                else:
                    self.plt.tight_layout('fit')

            # reset style
            self.bfit.draw_style.set(style)

        # reset asym mode
        self.bfit.fetch_files.asym_type.set(asym_mode_fetch)

    # ======================================================================= #
    def do_fit(self, *args):
        # fitter
        fitter = self.fitter
        figstyle = 'fit'

        # get fitter inputs
        fn_name = self.fit_function_title.get()
        ncomp = self.n_component.get()

        xlims = [self.xlo.get(), self.xhi.get()]
        if not xlims[0]:
            xlims[0] = '-inf'
            self.xlo.set('-inf')
        if not xlims[1]:
            xlims[1] = 'inf'
            self.xhi.set('inf')

        try:
            xlims = tuple(map(float, xlims))
        except ValueError as err:
            messagebox.showerror("Error", 'Bad input for xlims')
            self.logger.exception(str(err))
            raise err

        self.logger.info('Fitting with "%s" with %d components', fn_name, ncomp)

        # build data list
        data_list = []
        for key in self.fit_lines:

            # get fit line
            fitline = self.fit_lines[key]

            # bdata object
            bdfit = fitline.data

            # get entry values
            pdict = {}
            for line in fitline.lines:
                inpt = line.get('*')
                pdict[line.pname] = [inpt[k] for k in ('p0', 'blo', 'bhi', 'fixed', 'shared')]

            # doptions
            doptions = {}
            doptions['slr_bkgd_corr'] = self.bfit.correct_bkgd.get()

            if self.use_rebin.get():
                doptions['rebin'] = bdfit.rebin.get()

            if '1' in self.mode:
                dline = self.bfit.fetch_files.data_lines[key]
                doptions['omit'] = dline.bin_remove.get()
                if doptions['omit'] == dline.bin_remove_starter_line:
                    doptions['omit'] = ''
            elif '2' in self.mode:
                pass
            else:
                msg = 'Fitting mode %s not recognized' % self.mode
                self.logger.error(msg)
                raise RuntimeError(msg)

            # make data list
            data_list.append([bdfit, pdict, doptions])

        # call fitter with error message, potentially
        self.fit_input = (fn_name, ncomp, data_list)

        # set up queue
        que = Queue()

        def run_fit():
            try:
                # fit_output keyed as {run:[key/par/cov/chi/fnpointer]}
                fit_output = fitter(fn_name=fn_name,
                                    ncomp=ncomp,
                                    data_list=data_list,
                                    hist_select=self.bfit.hist_select,
                                    asym_mode=self.bfit.get_asym_mode(self),
                                    xlims=xlims)
            except Exception as errmsg:
                self.logger.exception('Fitting error')
                que.put(str(errmsg))
                raise errmsg from None

            que.put(fit_output)

        # log fitting
        for d in data_list:
            self.logger.info('Fitting run %s: %s', self.bfit.get_run_key(d[0]), d[1:])

        # start fit
        popup = popup_ongoing_process(self.bfit,
                    target = run_fit,
                    message="Fitting in progress...",
                    queue = que,
                    do_disable = lambda : self.input_enable_disable(self.fit_data_tab, state='disabled'),
                    do_enable = lambda : self.input_enable_disable(self.fit_data_tab, state='normal'),
                    )
        output = popup.run()

        # fit success
        if type(output) is tuple:
            fit_output, gchi = output

        # error message
        elif type(output) is str:
            messagebox.showerror("Error", output)
            return

        # fit cancelled
        elif output is None:
            return

        # get fit functions
        fns = fitter.get_fit_fn(fn_name, ncomp, data_list)

        # set output results
        for key, df in fit_output.items(): # iterate run ids

            # get fixed and shared
            fs = {'fixed':[], 'shared':[], 'parnames':[]}
            for line in self.fit_lines[key].lines:

                fs['parnames'].append(line.pname)
                fs['fixed'].append(line.get('fixed'))
                fs['shared'].append(line.get('shared'))

            df2 = pd.concat((df, pd.DataFrame(fs).set_index('parnames')), axis='columns')

            # make output
            new_output = {'results': df2,
                          'fn': fns[key],
                          'gchi': gchi}

            self.bfit.data[key].set_fitresult(new_output)
            self.bfit.data[key].fit_title = self.fit_function_title.get()
            self.bfit.data[key].ncomp = self.n_component.get()

        # display run results
        for key in self.fit_lines.keys():
            self.fit_lines[key].show_fit_result()

        # show global chi
        self.gchi_label['text'] = str(np.around(gchi, 2))

        self.do_end_of_fit()

    # ======================================================================= #
    def do_fit_model(self):

        self.logger.info('Launching fit model popup')

        if hasattr(self, 'pop_fitres'):
            p = self.pop_fitres

            # don't make more than one window
            if Toplevel.winfo_exists(p.win):
                p.win.lift()
                return

            # make a new window, using old inputs and outputs
            self.pop_fitres = popup_fit_results(self.bfit,
                                    input_fn_text=p.input_fn_text,
                                    output_par_text=p.output_par_text_val,
                                    output_text=p.output_text_val,
                                    chi=p.chi,
                                    x = p.xaxis.get(),
                                    y = p.yaxis.get())

        # make entirely new window
        else:
            self.pop_fitres = popup_fit_results(self.bfit)

    # ======================================================================= #
    def do_gui_param(self, id=''):
        """Set initial parmeters with GUI"""

        self.logger.info('Launching initial fit parameters popup')
        popup_param(self.bfit, id)

    # ======================================================================= #
    def do_set_result_as_initial(self, *args):
        """Set initial parmeters as the fitting results"""

        self.logger.info('Setting initial parameters as fit results')

        # set result to initial value
        for k in self.fit_lines.keys():

            # get line
            fline = self.fit_lines[k]

            # set parameters
            for line in fline.lines:
                val = line.get('res')
                line.set(p0=val)

    # ======================================================================= #
    def do_reset_initial(self, *args):
        """Reset initial parmeters to defaults"""

        self.logger.info('Reset initial parameters')

        # reset lines
        for fline in self.fit_lines.values():
            fline.get_new_parameters(force_modify=True)
            fitpar = fline.data.fitpar
            for line in fline.lines:
                values = {c: fitpar.loc[line.pname, c] for c in fitpar.columns}
                line.set(**values)

    # ======================================================================= #
    def draw_param(self, *_):
        """Draw the fit parameters"""
        figstyle = 'param'

        # get draw components
        xdraw = self.xaxis.get()
        ydraw = self.yaxis.get()
        ann = self.annotation.get()
        label = self.par_label.get()

        # check for bad all the above input
        if ydraw == xdraw == 'all the above':
            msg = 'x and y cannot both be "all the above"'
            self.logger.error(msg)
            messagebox.showerror('Error', msg)
            raise RuntimeError(msg)

        # check for all the above - draw recursively
        if ydraw == 'all the above' or xdraw == 'all the above':

            self.logger.info("Drawing param as 'all the above'")

            # check which one we're modifying
            do_x = xdraw == 'all the above'

            # draw in new windows
            draw_style = self.bfit.draw_style.get()
            self.bfit.draw_style.set('new')

            for val in self.yaxis_combobox['values']:
                if val == 'all the above':
                    break
                elif val != '':

                    # set axis value
                    if do_x:    self.xaxis.set(val)
                    else:       self.yaxis.set(val)

                    # draw
                    self.draw_param()

            # reset axis value
            if do_x:    self.xaxis.set('all the above')
            else:       self.yaxis.set('all the above')

            # reset draw style
            self.bfit.draw_style.set(draw_style)

            return

        self.logger.info('Draw fit parameters "%s" vs "%s" with annotation "%s"'+\
                         ' and label %s', ydraw, xdraw, ann, label)

        # get plottable data
        try:
            xvals, xerrs = self.get_values(xdraw)
            yvals, yerrs = self.get_values(ydraw)
        except UnboundLocalError as err:
            self.logger.error('Bad input parameter selection')
            messagebox.showerror("Error", 'Select two input parameters')
            raise err from None
        except (KeyError, AttributeError) as err:
            self.logger.error('Parameter "%s or "%s" not found for drawing',
                              xdraw, ydraw)
            messagebox.showerror("Error",
                    'Drawing parameter "%s" or "%s" not found' % (xdraw, ydraw))
            raise err from None

        # get asymmetric errors
        if type(xerrs) is tuple:
            xerrs_l = xerrs[0]
            xerrs_h = xerrs[1]
        else:
            xerrs_l = xerrs
            xerrs_h = xerrs

        if type(yerrs) is tuple:
            yerrs_l = yerrs[0]
            yerrs_h = yerrs[1]
        else:
            yerrs_l = yerrs
            yerrs_h = yerrs

        # get annotation
        if ann != '':
            try:
                ann, _ = self.get_values(ann)
            except UnboundLocalError:
                ann = None
            except (KeyError, AttributeError) as err:
                self.logger.error('Bad input annotation value "%s"', ann)
                messagebox.showerror("Error",
                        'Annotation "%s" not found' % (ann))
                raise err from None

        # fix annotation values (blank to none)
        else:
            ann = None

        # get mouseover annotation labels
        mouse_label, _ = self.get_values('Unique Id')

        # sort by x values
        idx = np.argsort(xvals)
        xvals = np.asarray(xvals)[idx]
        yvals = np.asarray(yvals)[idx]

        xerrs_l = np.asarray(xerrs_l)[idx]
        yerrs_l = np.asarray(yerrs_l)[idx]
        xerrs_h = np.asarray(xerrs_h)[idx]
        yerrs_h = np.asarray(yerrs_h)[idx]

        if ann is not None:
            ann = np.asarray(ann)[idx]

        mouse_label = np.asarray(mouse_label)[idx]

        # fix annotation values (round floats)
        if ann is not None:
            number_string = '%.'+'%df' % self.bfit.rounding
            for i, a in enumerate(ann):
                if type(a) in [float, np.float64]:
                    ann[i] = number_string % np.around(a, self.bfit.rounding)

        # get default data_id
        if label:
            draw_id = label
        else:
            draw_id = ''

            if self.bfit.draw_style.get() == 'stack':
                ax = self.plt.gca(figstyle)

        # make new window
        style = self.bfit.draw_style.get()
        if style == 'new' or not self.plt.active[figstyle]:
            self.plt.figure(figstyle)
        elif style == 'redraw':
            self.plt.clf(figstyle)

        # get axis
        ax = self.plt.gca(figstyle)

        # set dates axis
        if xdraw in ('Start Time', ):
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%y/%m/%d (%H:%M)'))
            xvals = np.array([datetime.datetime.fromtimestamp(x) for x in xvals])
            xerrs = None
            ax.tick_params(axis='x', which='major', labelsize='x-small')
        else:
            try:
                ax.get_xaxis().get_major_formatter().set_useOffset(False)
            except AttributeError:
                pass

        if ydraw in ('Start Time', ):
            ax.yaxis.set_major_formatter(mdates.DateFormatter('%y/%m/%d (%H:%M)'))
            yvals = mdates.epoch2num(yvals)
            yerrs = None
            ax.tick_params(axis='y', which='major', labelsize='x-small')
        else:
            try:
                ax.get_yaxis().get_major_formatter().set_useOffset(False)
            except AttributeError:
                pass

        # remove component label
        ncomp = self.n_component.get()
        xsuffix = ''
        ysuffix = ''

        if ncomp > 1:

            fn_params = self.fitter.gen_param_names(self.fit_function_title.get(), ncomp)

            if xdraw in fn_params or 'Beta-Avg 1/<T1>' in xdraw or 'T1' in xdraw:

                if '_' in xdraw:
                    spl = xdraw.split('_')
                    xdraw = '_'.join(spl[:-1])
                    xsuffix = ' [%s]' % spl[-1]
                else:
                    xsuffix = ''

            if ydraw in fn_params or 'Beta-Avg 1/<T1>' in ydraw or 'T1' in ydraw:

                if '_' in ydraw:
                    spl = ydraw.split('_')
                    ydraw = '_'.join(spl[:-1])
                    ysuffix = ' [%s]' % spl[-1]
                else:
                    ysuffix = ''

        # pretty labels
        xdraw = self.fitter.pretty_param.get(xdraw, xdraw)
        ydraw = self.fitter.pretty_param.get(ydraw, ydraw)

        # add suffix for multiple labels
        xdraw = xdraw + xsuffix
        ydraw = ydraw + ysuffix

        # attempt to insert units and scale
        unit_scale, unit = self.bfit.units.get(self.mode, [1, ''])
        if '%s' in xdraw:
            xdraw = xdraw % unit
            xvals *= unit_scale
            xerrs_h *= unit_scale
            xerrs_l *= unit_scale
        if '%s' in ydraw:
            ydraw = ydraw % unit
            yvals *= unit_scale
            yerrs_h *= unit_scale
            yerrs_l *= unit_scale

        # check for nan errors
        if all(np.isnan(xerrs_h)): xerrs_h = None
        if all(np.isnan(xerrs_l)): xerrs_l = None
        if all(np.isnan(yerrs_h)): yerrs_h = None
        if all(np.isnan(yerrs_l)): yerrs_l = None

        if xerrs_h is None and xerrs_l is None:     xerr = None
        else:                                       xerr = (xerrs_l, xerrs_h)
        if yerrs_h is None and yerrs_l is None:     yerr = None
        else:                                       yerr = (yerrs_l, yerrs_h)

        # draw
        f = self.plt.errorbar(  figstyle,
                                draw_id,
                                xvals,
                                yvals,
                                xerr = xerr,
                                yerr = yerr,
                                label=draw_id,
                                annot_label=mouse_label,
                                **self.bfit.style)
        self._annotate(draw_id, xvals, yvals, ann, color=f[0].get_color(), unique=False)

        # format date x axis
        if xerrs is None:   self.plt.gcf(figstyle).autofmt_xdate()

        # plot elements
        self.plt.xlabel(figstyle, xdraw)
        self.plt.ylabel(figstyle, ydraw)
        self.plt.tight_layout(figstyle)

        if draw_id:
            self.plt.legend(figstyle, fontsize='x-small')

        # bring window to front
        raise_window()

    # ======================================================================= #
    def export(self, savetofile=True, filename=None):
        """Export the fit parameter and file headers"""
        # get values and errors
        val = {}

        for v in self.xaxis_combobox['values']:
            if v == '': continue

            try:
                v2 = self.get_values(v)

            # value not found
            except (KeyError, AttributeError):
                continue

            # if other error, don't crash but print the result
            except Exception:
                traceback.print_exc()
            else:
                val[v] = v2[0]

                if type(v2[1]) is tuple:
                    val['Error- '+v] = v2[1][0]
                    val['Error+ '+v] = v2[1][1]
                else:
                    val['Error '+v] = v2[1]

        # get fixed and shared, if fitted
        keylist = []
        for k, line in self.fit_lines.items():
            keylist.append(k)
            data = line.data

            if not all(data.fitpar['res'].isna()):

                for kk in data.fitpar.index:

                    name = 'fixed '+kk
                    if name not in val.keys(): val[name] = []
                    val[name].append(data.fitpar.loc[kk, 'fixed'])

                    name = 'shared '+kk
                    if name not in val.keys(): val[name] = []
                    val[name].append(data.fitpar.loc[kk, 'shared'])

        # make data frame for output
        df = pd.DataFrame(val)
        df.set_index('Run Number', inplace=True)

        # drop completely empty columns
        bad_cols = [c for c in df.columns if all(df[c].isna())]
        for c in bad_cols:
            df.drop(c, axis='columns', inplace=True)

        if savetofile:

            # get file name
            if filename is None:
                filename = filedialog.asksaveasfilename(filetypes=[('csv', '*.csv'),
                                                                   ('allfiles', '*')],
                                                    defaultextension='.csv')
                if not filename:
                    return
            self.logger.info('Exporting parameters to "%s"', filename)

            # check extension
            if os.path.splitext(filename)[1] == '':
                filename += '.csv'

            # write header
            data = self.bfit.data[list(self.fit_lines.keys())[0]]

            if hasattr(data, 'fit_title'):
                header = ['# Fit function : %s' % data.fit_title,
                          '# Number of components: %d' % data.ncomp,
                          '# Global Chi-Squared: %s' % self.gchi_label['text']
                          ]

                # add constrained equations to header
                if self.pop_fitconstr.defined:
                    head2 = ['# ',
                             '# Constrained parameters',
                             ]
                    header.extend(head2)

                    for d, e in zip(self.pop_fitconstr.defined, self.pop_fitconstr.eqn):
                        header.append('# {defi} = {eqn}'.format(defi=d, eqn=e))

            else:
                header = []

            header.extend(['#\n# Generated by bfit v%s on %s' % (__version__, datetime.datetime.now()),
                          '#\n#\n'])

            with open(filename, 'w') as fid:
                fid.write('\n'.join(header))

            # write data
            df.to_csv(filename, mode='a+')
            self.logger.debug('Export success')
        else:
            self.logger.info('Returned exported parameters')
            return df

    # ======================================================================= #
    def export_fit(self, savetofile=True, directory=None):
        """Export the fit lines as csv files"""

        # filename
        filename = self.bfit.fileviewer.default_export_filename
        filename = '_fit'.join(os.path.splitext(filename))

        if directory is None:
            directory = filedialog.askdirectory()
            if not directory:
                return

        filename = os.path.join(directory, filename)

        # asymmetry type
        asym_mode = self.bfit.get_asym_mode(self)

        # get data and write
        for id in self.fit_lines.keys():

            # get data
            data = self.bfit.data[id]
            t, a, da = data.asym(asym_mode)

            # get fit data
            fitx = np.linspace(min(t), max(t), self.n_fitx_pts)

            try:
                fit_par = data.fitpar.loc[data.parnames, 'res']
            except AttributeError:
                continue
            dfit_par_l = data.fitpar.loc[data.parnames, 'dres-']
            dfit_par_h = data.fitpar.loc[data.parnames, 'dres+']
            fity = data.fitfn(fitx, *fit_par)

            if data.mode in self.bfit.units:
                unit = self.bfit.units[data.mode]
                fitxx = fitx*unit[0]
                xlabel = self.bfit.xlabel_dict[self.mode] % unit[1]
            else:
                fitxx = fitx
                xlabel = self.bfit.xlabel_dict[self.mode]

            # write header
            fname = filename%(data.year, data.run)
            header = ['# %s' % data.id,
                      '# %s' % data.title,
                      '# Fit function : %s' % data.fit_title,
                      '# Number of components: %d' % data.ncomp,
                      '# Rebin: %d' % data.rebin.get(),
                      '# Bin Omission: %s' % data.omit.get().replace(
                                self.bfit.fetch_files.bin_remove_starter_line, ''),
                      '# Chi-Squared: %f' % data.chi,
                      '# Parameter names: %s' % ', '.join(data.parnames),
                      '# Parameter values: %s' % ', '.join(list(map(str, fit_par))),
                      '# Parameter errors (-): %s' % ', '.join(list(map(str, dfit_par_l))),
                      '# Parameter errors (+): %s' % ', '.join(list(map(str, dfit_par_h))),
                      '#',
                      '# Generated by bfit v%s on %s' % (__version__, datetime.datetime.now()),
                      '#']

            with open(fname, 'w') as fid:
                fid.write('\n'.join(header) + '\n')

            # write data
            df = pd.DataFrame({xlabel:fitx, 'asymmetry':fity})
            df.to_csv(fname, index=False, mode='a+')
            self.logger.info('Exporting fit to %s', fname)

    # ======================================================================= #
    def get_values(self, select):
        """ Get plottable values from all runs"""

        data = self.bfit.data
        dlines = self.bfit.fetch_files.data_lines

        # draw only selected runs
        runs = [dlines[k].id for k in dlines if dlines[k].check_state.get()]
        runs.sort()

        self.logger.debug('Fetching parameter %s', select)

        # get values
        out = np.array([data[r].get_values(select) for r in runs], dtype=object)

        val = out[:, 0]
        err = np.array(out[:, 1].tolist())
        if len(err.shape) > 1:
            err = (err[:, 0], err[:, 1])

        return (val, err)

    # ======================================================================= #
    def input_enable_disable(self, parent, state, first=True):
        """
            Prevent input while fitting by disabling options

            state: "disabled" or "normal"
            first: do non-recursive items (i.e. menus, tabs)
        """

        if first:

            # disable tabs
            self.bfit.notebook.tab(1, state=state)

            # disable menu options
            file = self.bfit.menus['File']
            file.entryconfig("Run Commands", state=state)
            file.entryconfig("Export Fits", state=state)
            file.entryconfig("Save State", state=state)
            file.entryconfig("Load State", state=state)

            settings = self.bfit.menus['Settings']
            settings.entryconfig("Probe Species", state=state)

            draw_mode = self.bfit.menus['Draw Mode']
            draw_mode.entryconfig("Use NBM in asymmetry", state=state)
            draw_mode.entryconfig("Draw 1f/1x as PPM shift", state=state)

            self.bfit.menus['menubar'].entryconfig("Minimizer", state=state)

        # disable everything in fit_tab
        for child in parent.winfo_children():

            # exceptions
            if child in (self.xaxis_combobox,
                         self.yaxis_combobox,
                         self.annotation_combobox,
                         self.par_label_entry):
                continue


            # disable
            try:
                if state == 'disabled':
                    child.old_state = child['state']
                    child.configure(state=state)
                else:
                    child.configure(state=child.old_state)
            except (TclError, AttributeError):
                pass
            self.input_enable_disable(child, state=state, first=False)

    # ======================================================================= #
    def return_binder(self):
        """
            Binding to entery key press, depending on focus.

            FOCUS                   ACTION

            comboboxes or buttons   draw_param
                in right frame
            else                    do_fit
        """

        # get focus
        focus = self.bfit.root.focus_get()

        # right frame items
        draw_par_items = (  self.xaxis_combobox,
                            self.yaxis_combobox,
                            self.annotation_combobox,
                            self.par_label_entry)

        # do action
        if focus in draw_par_items:
            self.draw_param()
        elif focus == self.n_component_box:
            self.populate_param(force_modify=True)
        elif focus == self.bfit.root:
            pass
        else:
            self.do_fit()

    # ======================================================================= #
    def set_lines(self, pname, col, value, skipline=None):
        """
            Modify all input fields of each line to match the altered one
            conditional on self.set_as_group

            pname: string, parameter being changed
            col:   str, column being changed
            value: new value to assign
            skipline: if this line, don't modify
        """

        for fitline in self.fit_lines.values():

            # trivial case
            if not fitline.lines:
                return

            # get line id
            id = [line.pname for line in fitline.lines].index(pname)
            line = fitline.lines[id]

            if line == skipline:
                continue

            # set
            line.set(**{col:value})

    # ======================================================================= #
    def show_all_results(self):
        """Make a window to display table of fit results"""

        self.logger.info('Launching parameter table popup')

        # get fit results
        df = self.export(savetofile=False)
        popup_show_param(df)

    # ======================================================================= #
    def show_constr_window(self):

        self.logger.info('Launching fit constraints popup')

        p = self.pop_fitconstr

        # don't make more than one window
        if hasattr(p, 'win') and Toplevel.winfo_exists(p.win):
            p.win.lift()

        # make a new window, using old inputs and outputs
        else:
            p.show()

    # ======================================================================= #
    def update_param(self, *args):
        """Update all figures with parameters drawn with new fit results"""

        self.logger.info('Updating parameter figures')

        # get list of figure numbers for parameters
        figlist = self.plt.plots['param']

        # set style to redraw
        current_active = self.plt.active['param']
        current_style = self.bfit.draw_style.get()
        self.bfit.draw_style.set('stack')

        # get current labels
        current_xlab = self.xaxis.get()
        current_ylab = self.yaxis.get()

        # get current unit
        unit = self.bfit.units[self.mode]

        # back-translate pretty labels to originals
        ivd = {}
        for  k, v in self.fitter.pretty_param.items():

            try:
                v = v % unit[1]
            except TypeError:
                pass

            ivd[v] = k

        for fig_num in figlist:

            # get figure and drawn axes
            ax = plt.figure(fig_num).axes[0]
            xlab = ax.get_xlabel()
            ylab = ax.get_ylabel()

            # remove multi-compoent extension
            try:
                ext_x = xlab.split('[')[1]
            except IndexError:
                ext_x = ''
            else:
                ext_x = ext_x.split(']')[0]
            xlab = xlab.split('[')[0].strip()

            try:
                ext_y = ylab.split('[')[1]
            except IndexError:
                ext_y = ''
            else:
                ext_y = ext_y.split(']')[0]
            ylab = ylab.split('[')[0].strip()

            # convert from fancy label to simple label
            xlab = ivd.get(xlab, xlab)
            ylab = ivd.get(ylab, ylab)

            # add multi-componet stuff
            if ext_x: xlab += '_%s' % ext_x
            if ext_y: ylab += '_%s' % ext_y

            # set new labels for drawing
            self.xaxis.set(xlab)
            self.yaxis.set(ylab)

            # draw new labels
            self.plt.active['param'] = fig_num
            self.draw_param()

            self.logger.debug('Updated figure %d (%s vs %s)', fig_num, ylab, xlab)

        # reset to old settings
        self.bfit.draw_style.set(current_style)
        self.xaxis.set(current_xlab)
        self.yaxis.set(current_ylab)
        self.plt.active['param'] = current_active
        plt.figure(current_active)



# Model the fit results with a function
# Derek Fujimoto
# August 2019

from tkinter import *
from tkinter import ttk, messagebox
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from bfit.gui.template_fit_popup import template_fit_popup
from bfit.backend.raise_window import raise_window
from bfit.fitting.minuit import minuit

# ========================================================================== #
class popup_fit_results(template_fit_popup):
    """
        Popup window for modelling the fit results with a function

        chi_label:      Label, chisquared output
        fittab:         notebook tab
        new_par:        dataframe, index: parnames, columns: p0, blo, bhi, res, err

        output_par_text_val string, contents of output_par_text
        output_text_val     dict of strings, contents of output_text

        par:            list, fit results
        reserved_pars:  dict, keys: x, y vals: strings of parameter names

        text:           string, model text

        xaxis:          StringVar, x axis drawing/fitting parameter
        yaxis:          StringVar, y axis drawing/fitting parameter

    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}

    window_title = 'Fit the results with a model'
    reserved_pars = ['x', 'y']

    # ====================================================================== #
    def __init__(self, bfit, input_fn_text='', output_par_text='', output_text='',
                 chi=np.nan, x='', y=''):

        super().__init__(bfit, input_fn_text)
        super().show()

        self.fittab = bfit.fit_files
        self.output_par_text_val = output_par_text

        if not output_text:
            self.output_text_val = {}
        else:
            self.output_text_val = output_text

        # text for output
        output_frame = ttk.Frame(self.right_frame, relief='sunken', pad=5)
        output_head1_label = ttk.Label(output_frame, text='Par Name')
        output_head2_label = ttk.Label(output_frame, text='p0')
        output_head3_label = ttk.Label(output_frame, text='Bounds')
        output_head4_label = ttk.Label(output_frame, text='Result')
        output_head5_label = ttk.Label(output_frame, text='Error (-)')
        output_head6_label = ttk.Label(output_frame, text='Error (+)')
        self.output_par_text = Text(output_frame, width=8, height=8)
        self.output_text = {k:Text(output_frame, width=8, height=8, wrap='none')\
                            for k in ('p0', 'blo', 'bhi', 'res', 'err-', 'err+')}

        # default starter strings
        if self.output_par_text_val:
            self.output_par_text.insert('1.0', self.output_par_text_val)
        self.output_par_text.config(state='disabled')

        if self.output_text_val:
            for k in self.output_text_val:
                self.output_text[k].insert('1.0', self.output_text_val[k])

        # disable results
        for k in ('res', 'err-', 'err+'):
            self.output_text[k].config(state='disabled', width=12)

        # key bindings and scrollbar
        scrollb_out = Scrollbar(output_frame, command=self.yview)
        self.output_par_text['yscrollcommand'] = scrollb_out.set
        for k in self.output_text:
            self.output_text[k].bind('<KeyRelease>', self.get_result_input)
            self.output_text[k]['yscrollcommand'] = scrollb_out.set

        c = 0; r = 0;
        output_head1_label.grid(column=c, row=r);        c+=1;
        output_head2_label.grid(column=c, row=r);        c+=1;
        output_head3_label.grid(column=c, row=r,
                                columnspan=2);          c+=2;
        output_head4_label.grid(column=c, row=r);        c+=1;
        output_head5_label.grid(column=c, row=r);        c+=1;
        output_head6_label.grid(column=c, row=r);        c+=1;

        c = 0; r += 1;
        self.output_par_text.grid(column=c, row=r, sticky=N); c+=1;
        for k in ('p0', 'blo', 'bhi', 'res', 'err-', 'err+'):
            self.output_text[k].grid(column=c, row=r, sticky=N); c+=1;

        # fitting button
        fit_button = ttk.Button(self.right_frame, text='Fit', command=self.do_fit)


        # initialize
        self.new_par = pd.DataFrame(columns=['name', 'p0', 'blo', 'bhi', 'res', 'err-', 'err+'])
        self.fittab = self.bfit.fit_files
        self.chi = chi

        # menus for x and y values
        axis_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)

        ttk.Label(  axis_frame,
                    text='Variable definitions:\n',
                    justify=LEFT).grid(column=0, row=0, columnspan=2, sticky=W)
        ttk.Label(axis_frame, text="x axis:").grid(column=0, row=1)
        ttk.Label(axis_frame, text="y axis:").grid(column=0, row=2)
        ttk.Label(axis_frame, text=' ').grid(column=0, row=3)

        self.xaxis = StringVar()
        self.yaxis = StringVar()

        if x:   self.xaxis.set(x)
        else:   self.xaxis.set(self.fittab.xaxis.get())

        if y:   self.yaxis.set(y)
        else:   self.yaxis.set(self.fittab.yaxis.get())

        self.xaxis_combobox = ttk.Combobox(axis_frame, textvariable=self.xaxis,
                                      state='readonly', width=19)
        self.yaxis_combobox = ttk.Combobox(axis_frame, textvariable=self.yaxis,
                                      state='readonly', width=19)

        self.xaxis_combobox['values'] = self.fittab.xaxis_combobox['values']
        self.yaxis_combobox['values'] = self.fittab.yaxis_combobox['values']

        # module names
        module_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved module names:\n\n'

        maxk = max(list(map(len, list(self.modules.keys()))))

        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in self.modules.items()])
        s += '\n'
        modules_label = ttk.Label(module_frame, text=s, justify=LEFT)

        # chisquared output
        chi_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        self.chi_label = ttk.Label(chi_frame,
                                    text='ChiSq: %.2f' % np.around(chi, 2),
                                    justify=LEFT)

        # Text entry
        self.entry_label['text'] = 'Enter a one line equation using "x"'+\
                                 ' to model y(x)'+\
                                 '\nEx: "y = a*x+b"'

        # draw button
        button_draw = ttk.Button(self.right_frame, text='Draw',
                                 command=self.draw_model, pad=1)

        # gridding
        modules_label.grid(column=0, row=0)
        self.chi_label.grid(column=0, row=0)
        self.xaxis_combobox.grid(column=1, row=1)
        self.yaxis_combobox.grid(column=1, row=2)

        axis_frame.grid(column=0, row=0, rowspan=1, sticky=(E, W), padx=1, pady=1)
        module_frame.grid(column=0, row=1, sticky=(E, W), padx=1, pady=1)
        chi_frame.grid(column=0, row=2, sticky=(E, W), padx=1, pady=1, rowspan=2)

        button_draw.grid(column=0, row=3, sticky=(E, W))

        # gridding
        scrollb_out.grid(row=r, column=c, sticky='nsew')
        output_frame.grid(column=0, row=1, sticky=(N, E, W, S), padx=1, pady=1)
        fit_button.grid(column=0, row=2, sticky=(N, E, W), padx=1, pady=1)

    # ====================================================================== #
    def _do_fit(self, text):

        # save model text
        self.text= text[-1]

        # get fit data
        xstr = self.xaxis.get()
        ystr = self.yaxis.get()

        # Make model
        parnames = self.output_par_text.get('1.0', END).split('\n')[:-1]
        parstr = ', '.join(parnames)
        eqn = text[-1].split('=')[-1]
        model = 'lambda x, %s : %s' % (parstr, eqn)

        self.logger.info('Fitting model %s for x="%s", y="%s"', model, xstr, ystr)
        self.model_fn = eval(model)
        npar = len(parnames)

        # set up p0, bounds
        p0 = self.new_par['p0'].values
        blo = self.new_par['blo'].values
        bhi = self.new_par['bhi'].values

        p0 = list(map(float, p0))
        blo = list(map(float, blo))
        bhi = list(map(float, bhi))

        # get data to fit
        xvals, xerrs_l, xerrs_h = self._get_data(xstr)
        yvals, yerrs_l, yerrs_h = self._get_data(ystr)

        # minimize
        m = minuit(self.model_fn, xvals, yvals,
                  dy = yerrs_h,
                  dx = xerrs_h,
                  dy_low = yerrs_l,
                  dx_low = xerrs_l,
                  name = parnames,
                  print_level = 0,
                  limit = np.array([blo, bhi]).T,
                  start = p0,
                  )

        m.migrad()
        m.hesse()
        m.minos()

        # print fitting quality
        try:
            print(m.fmin)
            print(m.params)
            print(m.merrors)
        except UnicodeEncodeError:
            pass

        # get results
        par = m.values
        self.par = par
        n = len(m.merrors)

        if npar == 1:
            std_l = np.array([m.merrors[i].lower for i in range(n)])
            std_h = np.array([std_l])
        else:
            std_l = np.array([m.merrors[i].lower for i in range(n)])
            std_h = np.array([m.merrors[i].upper for i in range(n)])

        # chi2
        chi = m.chi2
        self.chi_label['text'] = 'ChiSq: %.2f' % np.around(chi, 2)
        self.chi = chi

        self.logger.info('Fit model results: %s, Errors-: %s, Errors+: %s',
                        str(par), str(std_l), str(std_h))

        self.bfit.plt.figure('param')
        self.draw_data()
        self.draw_model()

        return (par, std_l, std_h)

    # ====================================================================== #
    def _get_data(self, xstr):
        """
            Get data and clean

            xstr: label for data to fetch
        """

        # get data
        try:
            xvals, xerrs = self.fittab.get_values(xstr)
        except UnboundLocalError as err:
            self.logger.error('Bad input parameter')
            messagebox.showerror("Error", 'Bad input parameter')
            raise err
        except (KeyError, AttributeError) as err:
            self.logger.error('Parameter "%s not found for fitting', xstr)
            messagebox.showerror("Error", 'Parameter "%s" not found' % xstr)
            raise err

        # split errors
        if type(xerrs) is tuple:
            xerrs_l = xerrs[0]
            xerrs_h = xerrs[1]
        else:
            xerrs_l = xerrs
            xerrs_h = xerrs

        xvals = np.asarray(xvals)
        xerrs_l = np.asarray(xerrs_l)
        xerrs_h = np.asarray(xerrs_h)

        # check errors
        if all(np.isnan(xerrs_l)): xerrs_l = None
        if all(np.isnan(xerrs_h)): xerrs_h = None

        return (xvals, xerrs_l, xerrs_h)

    # ====================================================================== #
    def do_after_parse(self, defined, eqn, new_par):

        # add result
        try:
            new_par = np.unique(np.concatenate(new_par))
        except TypeError:
            return

        for k in new_par:

            # bad input
            if not k: continue

            # set defaults
            if k not in self.new_par['name'].values:
                new_par2 = pd.DataFrame({'name':k, **self.default_parvals}, index=[0])

                if self.new_par.empty:
                    self.new_par = new_par2.copy()
                else:
                    self.new_par = pd.concat((self.new_par, new_par2),
                                             axis='index',
                                             ignore_index=True)

        # drop results
        for i, k in zip(self.new_par.index, self.new_par['name']):
            if k not in new_par:
                self.new_par.drop(i, inplace=True)

        # set fields
        self.new_par.sort_values('name', inplace=True)
        self.set_par_text()

    # ====================================================================== #
    def do_fit(self, *args):
        """
            Set up the fit functions and do the fit.
            Then map the outputs to the proper displays.
        """

        # parse text
        self.do_parse()

        # clean input
        text = self.input_fn_text.split('\n')
        text = [t.strip() for t in text if '=' in t]

        # check for no input
        if not text:    return

        try:
            # do the fit
            out = self._do_fit(text)
        except Exception as errmsg:
            raise errmsg from None

        # check if fit was success
        if out is None:
            return
        else:
            par, std_l, std_h  = out

        # display output for global parameters
        for i, j in enumerate(self.new_par.index):
            self.new_par.loc[j, 'res'] = par[i]
            self.new_par.loc[j, 'err-'] = std_l[i]
            self.new_par.loc[j, 'err+'] = std_h[i]
        self.set_par_text()

    # ======================================================================= #
    def draw_data(self):
        figstyle = 'param'

        # get draw components
        xstr = self.xaxis.get()
        ystr = self.yaxis.get()
        id = self.fittab.par_label.get()

        self.logger.info('Draw model fit data "%s" vs "%s"', ystr, xstr)

        # get data
        xvals, xerrs_l, xerrs_h = self._get_data(xstr)
        yvals, yerrs_l, yerrs_h = self._get_data(ystr)

        # sort by x values, check for empty arrays
        idx = np.argsort(xvals)
        xvals = np.asarray(xvals)[idx]
        yvals = np.asarray(yvals)[idx]

        if xerrs_h is not None:     xerrs_h = np.asarray(xerrs_h)[idx]
        if xerrs_l is not None:     xerrs_l = np.asarray(xerrs_l)[idx]
        if yerrs_h is not None:     yerrs_h = np.asarray(yerrs_h)[idx]
        if yerrs_l is not None:     yerrs_l = np.asarray(yerrs_l)[idx]

        if xerrs_h is None and xerrs_l is None:     xerrs = None
        else:                                       xerrs = (xerrs_l, xerrs_h)
        if yerrs_h is None and yerrs_l is None:     yerrs = None
        else:                                       yerrs = (yerrs_l, yerrs_h)

        # get mouseover annotation labels
        mouse_label, _ = self.fittab.get_values('Unique Id')
        mouse_label = np.asarray(mouse_label)[idx]

        #draw data
        self.bfit.plt.errorbar(figstyle, id, xvals, yvals,
                                 yerr=yerrs,
                                 xerr=xerrs,
                                 fmt='.',
                                 annot_label=mouse_label)

        # pretty labels
        xstr = self.fittab.fitter.pretty_param.get(xstr, xstr)
        ystr = self.fittab.fitter.pretty_param.get(ystr, ystr)

        # plot elements
        self.bfit.plt.xlabel(figstyle, xstr)
        self.bfit.plt.ylabel(figstyle, ystr)
        self.bfit.plt.tight_layout(figstyle)

        raise_window()

    # ======================================================================= #
    def draw_model(self):
        """Draw model line as stacked"""
        figstyle = 'param'

        self.logger.info('Draw model "%s"', self.text)

        # get fit function and label id
        fn = self.model_fn
        draw_id = self.fittab.par_label.get()

        # get default data_id
        if not draw_id and self.bfit.draw_style.get() == 'stack':
            ax = self.bfit.plt.gca(figstyle)

        # get x data
        xstr = self.xaxis.get()
        xvals, _, _ = self._get_data(xstr)

        # draw fit
        self.bfit.plt.gca('param')
        fitx = np.linspace(min(xvals), max(xvals), self.fittab.n_fitx_pts)
        f = self.bfit.plt.plot(figstyle, draw_id, fitx, fn(fitx, *self.par),
                               color='k', label=self.text, unique=False)
        plt.show()
        raise_window()

    # ====================================================================== #
    def get_result_input(self, *args):
        """
            Set new_par row to match changes made by user
        """

        # get text
        try:
            text = {k : list(map(float, self.output_text[k].get('1.0', END).split('\n')[:-1])) \
                for k in self.output_text}
        # no update if blank
        except ValueError:
            return

        # dataframe it
        try:
            text = pd.DataFrame(text)
        # bad input
        except ValueError:
            return

        # get names of the parameters
        parnames = self.output_par_text.get('1.0', END).split('\n')[:-1]

        # update
        par = self.new_par.set_index('name')
        for i, name in enumerate(parnames):
            par.loc[name] = text.iloc[i]
        par.reset_index(inplace=True)
        self.new_par = par
        self.logger.debug('get_result_input: updated new_par')

    # ====================================================================== #
    def set_par_text(self):
        """
            Set the textboxes based on stored results in self.newpar
        """

        # get strings
        set_par = self.new_par.astype(str)

        # round
        numstr = '%'+('.%df' % self.bfit.rounding)
        for k in ('res', 'err-', 'err+'):
            set_par[k] = set_par.loc[:, k].apply(\
                    lambda x : numstr % np.around(float(x), self.bfit.rounding))

        # enable setting
        for k in ('res', 'err-', 'err+'):
            self.output_text[k].config(state='normal')
        self.output_par_text.config(state='normal')

        self.output_par_text.delete('1.0', END)
        self.output_par_text.insert(1.0, '\n'.join(set_par['name']))
        self.output_par_text_val = '\n'.join(set_par['name'])

        for k in self.output_text:
            self.output_text[k].delete('1.0', END)
            self.output_text[k].insert(1.0, '\n'.join(set_par[k]))
            self.output_text_val[k] = '\n'.join(set_par[k])

        # disable setting
        for k in ('res', 'err-', 'err+'):
            self.output_text[k].config(state='disabled')
        self.output_par_text.config(state='disabled')





# Test deadtime corrections popup
from numpy.testing import *
import numpy as np
import pandas as pd
import bdata as bd
from bfit.gui.bfit import bfit
from bfit.gui.popup_deadtime import popup_deadtime

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fileviewer
        tab2 = b.fetch_files
        
        # get data
        tab.year.set(2020)
        tab.runn.set(40123)
        tab.get_data()
            
        tab2.year.set(2020)
        tab2.run.set('40123 40127')
            
        # test
        try:
            return function(*args, **kwargs, b=b)
        finally:
            b.on_closing()
            del b
            
    return wrapper

@with_bfit
def test_default_run_set(b=None):
    
    dead = popup_deadtime(b)
    assert_equal(dead.year.get(), b.fileviewer.data.year, err_msg='Year not copied')
    assert_equal(dead.run.get(), b.fileviewer.data.run, err_msg='Run not copied')

@with_bfit
def test_find_free(b=None):
    
    dead = popup_deadtime(b)
    
    # both free
    dead.fix_dt.set(False)
    dead.fix_c.set(False)
    dead.find()
    
    assert dead.dt != 0
    assert dead.c != 1

@with_bfit
def test_find_fixed_c(b=None):
    
    dead = popup_deadtime(b)
    
    # fixed helicity
    dead.fix_dt.set(False)
    dead.fix_c.set(True)
    dead.find()
    
    assert dead.dt != 0
    assert dead.c == 1

@with_bfit
def test_find_fixed_dt(b=None):
    
    dead = popup_deadtime(b)
    
    # fixed deadtime
    dead.fix_dt.set(True)
    dead.fix_c.set(False)
    dead.find()
    
    assert dead.dt == 0
    assert dead.c != 1
    
@with_bfit
def test_find_fixed_both(b=None):
    
    dead = popup_deadtime(b)
    
    # fixed deadtime
    dead.fix_dt.set(True)
    dead.fix_c.set(True)
    dead.find()
    
    assert dead.dt == 0
    assert dead.c == 1
    
@with_bfit
def test_draw(b=None):
    
    dead = popup_deadtime(b)
    dead.find()
    dead.draw()

@with_bfit
def test_activate(b=None):
    
    dead = popup_deadtime(b)

    # activate off
    assert b.deadtime_switch.get() == False
    assert_equal(b.deadtime, 0, "Global deadtime initial state")
    
    # find to activate
    dead.find()
    assert b.deadtime_switch.get() == True
